# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""CLI-related utilities for Charmcraft."""
import datetime
import enum
import json
import numbers
import sys
from collections.abc import Collection, Iterable
from dataclasses import dataclass
from typing import Literal, overload

import tabulate
from craft_cli import emit

from charmcraft.env import is_charmcraft_running_in_managed_mode


class SingleOptionEnsurer:
    """Argparse helper to ensure that the option is specified only once, converting it properly.

    Receives a callable to convert the string from command line to the desired object.

    Example of use:

        parser.add_argument('-n', '--number',  type=SingleOptionEnsurer(int), required=True)

    No lower limit is checked, that is verified with required=True in the argparse definition.
    """

    def __init__(self, converter):
        self.converter = converter
        self.count = 0

    def __call__(self, value):
        """Run by argparse to validate and convert the given argument."""
        self.count += 1
        if self.count > 1:
            raise ValueError("the option can be specified only once")
        return self.converter(value)


@dataclass(frozen=True)
class ResourceOption:
    """Argparse helper to validate and convert a 'resource' option.

    Receives a callable to convert the string from command line to the desired object.

    Example of use:

        parser.add_argument('--resource',  type=ResourceOption())
    """

    name: str | None = None
    revision: int | None = None

    def __call__(self, value):
        """Run by argparse to validate and convert the given argument."""
        parts = [x.strip() for x in value.split(":")]
        parts = [p for p in parts if p]
        if len(parts) == 2:
            name, revision = parts
            try:
                revision = int(revision)
            except ValueError:
                pass
            else:
                if revision >= 0:
                    return ResourceOption(name, revision)
        msg = (
            "the resource format must be <name>:<revision> (revision being a non-negative integer)"
        )
        raise ValueError(msg)


@dataclass(frozen=True)
class ChoicesList:
    """Argparse helper to make a list argument from a collection of choices.

    For example, a list of digits might look like:

        parser.add_argument("--digits", type=ChoicesList(string.digits))
    """

    choices: Collection

    def __call__(self, value: str) -> list[str]:
        """Validate and get the chosen list of choices.

        :param value: The list as a comma-separated string from the CLI
        :returns: A list of string values.
        :raises ValueError: if any invalid choices are in the list.
        """
        values = value.split(",")
        invalid_values = set(values) - set(self.choices)
        if invalid_values:
            raise ValueError(f"invalid values: {', '.join(invalid_values)}")
        return values


def confirm_with_user(prompt: str, default: bool = False) -> bool:
    """Query user for yes/no answer.

    If stdin is not a tty, the default value is returned.

    If user returns an empty answer, the default value is returned.

    :returns: True if answer starts with [yY], False if answer starts with [nN],
        otherwise the default.
    """
    if is_charmcraft_running_in_managed_mode():
        raise RuntimeError("confirmation not yet supported in managed-mode")

    if not sys.stdin.isatty():
        return default

    choices = " [Y/n]: " if default else " [y/N]: "

    with emit.pause():
        reply = input(prompt + choices).lower().strip()

    if reply and reply[0] == "y":
        return True
    elif reply and reply[0] == "n":
        return False
    else:
        return default


def humanize_list(items: Iterable[str], conjunction: str) -> str:
    """Format a list into a human-readable string.

    :param items: list to humanize, must not be empty
    :param conjunction: the conjunction used to join the final element to
                        the rest of the list (e.g. 'and').
    """
    if not items:
        raise ValueError("Cannot humanize an empty list.")
    *initials, final = map(repr, sorted(items))
    if not initials:
        return final
    return f"{', '.join(initials)} {conjunction} {final}"


def format_timestamp(dt: datetime.datetime) -> str:
    """Convert a datetime object (with or without timezone) to a string.

    The format is

        <DATE>T<TIME>Z

    Always in UTC.
    """
    if dt.tzinfo is not None and dt.tzinfo.utcoffset(None) is not None:
        # timezone aware
        dtz = dt.astimezone(datetime.timezone.utc)
    else:
        # timezone naive, assume it's UTC
        dtz = dt
    return dtz.strftime("%Y-%m-%dT%H:%M:%SZ")


class OutputFormat(enum.Enum):
    """Output format options for commands."""

    DEFAULT = None
    JSON = "json"
    TABLE = "table"


@overload
def format_content(content: dict[str, str], fmt: Literal[OutputFormat.TABLE, "table"]) -> str: ...


@overload
def format_content(
    content: str | (numbers.Real | (list | dict)), fmt: OutputFormat | (str | None)
) -> str: ...


def format_content(content, fmt=None):
    """Format command output."""
    if not isinstance(fmt, OutputFormat):
        try:
            fmt = OutputFormat(fmt)
        except ValueError:
            raise ValueError(f"Unknown output format {str(fmt)}")

    if fmt == OutputFormat.JSON:
        return json.dumps(content, indent=4)
    if fmt == OutputFormat.TABLE:
        return tabulate.tabulate(content, headers="keys")
    return str(content)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for the config model."""
import math

import pydantic
import pytest

from charmcraft.models.config import (
    JujuBooleanOption,
    JujuConfig,
    JujuFloatOption,
    JujuIntOption,
    JujuStringOption,
)


@pytest.mark.parametrize(
    "options",
    [
        None,
        {},
        {
            "favourite integer": {"type": "int"},
            "favourite number": {"type": "float", "default": math.pi},
            "catchphrase": {"type": "string", "description": "What's your catchphrase?"},
            "default_answer": {
                "type": "boolean",
                "description": "Yes/no true or false",
                "default": True,
            },
        },
    ],
)
def test_valid_config(options):
    assert JujuConfig.model_validate({"options": options}) == JujuConfig(options=options)


def test_empty_config():
    JujuConfig.model_validate({})


@pytest.mark.parametrize(
    ("option", "type_"),
    [
        ({"type": "int"}, JujuIntOption),
        ({"type": "float"}, JujuFloatOption),
        ({"type": "string"}, JujuStringOption),
        ({"type": "boolean"}, JujuBooleanOption),
        ({"type": "float", "default": 0}, JujuFloatOption),
    ],
)
def test_correct_option_type(option, type_):
    config = JujuConfig(options={"my-opt": option})

    assert isinstance(config.options["my-opt"], type_)


@pytest.mark.parametrize(
    ("option", "match"),
    [
        (None, "Input should be a valid dict"),
        ({}, "Unable to extract tag using discriminator 'type'"),
        (
            {"type": "stargate"},
            "Input tag 'stargate' found using 'type' does not match any of the expected tags:",
        ),
        ({"type": "int", "default": 3.14}, "Input should be a valid integer"),
        ({"type": "float", "default": "pi"}, "Input should be a valid number"),
        ({"type": "boolean", "default": "maybe"}, "Input should be a valid boolean"),
    ],
)
def test_invalid_options(option, match):
    with pytest.raises(pydantic.ValidationError, match=match):
        JujuConfig(options={"my-opt": option})



# Copyright 2021-2024 Canonical Ltd.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License version 3 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Charmcraft's reactive plugin for craft-parts."""

import json
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Literal, cast

import overrides
from craft_parts import plugins
from craft_parts.errors import PluginEnvironmentValidationError


class ReactivePluginProperties(plugins.PluginProperties, frozen=True):
    """Properties used to pack reactive charms using charm-tools."""

    plugin: Literal["reactive"] = "reactive"
    source: str = "."
    reactive_charm_build_arguments: list[str] = []


class ReactivePluginEnvironmentValidator(plugins.validator.PluginEnvironmentValidator):
    """Check the execution environment for the Reactive plugin.

    :param part_name: The part whose build environment is being validated.
    :param env: A string containing the build step environment setup.
    """

    def validate_environment(self, *, part_dependencies: list[str] | None = None):
        """Ensure the environment contains dependencies needed by the plugin.

        :param part_dependencies: A list of the parts this part depends on.

        :raises PluginEnvironmentValidationError: If the environment is invalid.
        """
        try:
            version_data = json.loads(self._execute("charm version --format json"))

            tool_name = "charm-tools"
            if not (
                tool_name in version_data
                and "version" in version_data[tool_name]
                and "git" in version_data[tool_name]
            ):
                raise PluginEnvironmentValidationError(
                    part_name=self._part_name,
                    reason=f"invalid charm tools version {version_data}",
                )
            tools_version = (
                f"{tool_name} {version_data[tool_name]['version']} "
                f"({version_data[tool_name]['git']})"
            )
            print(f"found {tools_version}")
        except ValueError as err:
            raise PluginEnvironmentValidationError(
                part_name=self._part_name,
                reason="invalid charm tools installed",
            ) from err
        except subprocess.CalledProcessError as err:
            if err.returncode != plugins.validator.COMMAND_NOT_FOUND:
                raise PluginEnvironmentValidationError(
                    part_name=self._part_name,
                    reason=f"charm tools failed with error code {err.returncode}",
                ) from err

            if part_dependencies is None or "charm-tools" not in part_dependencies:
                raise PluginEnvironmentValidationError(
                    part_name=self._part_name,
                    reason=(
                        f"charm tool not found and part {self._part_name!r} "
                        f"does not depend on a part named 'charm-tools'"
                    ),
                ) from err


class ReactivePlugin(plugins.Plugin):
    """Build a reactive charm using charm-tools."""

    properties_class = ReactivePluginProperties
    validator_class = ReactivePluginEnvironmentValidator

    @overrides.override
    def get_build_snaps(cls) -> set[str]:
        """Return a set of required snaps to install in the build environment."""
        return set()

    def get_build_packages(self) -> set[str]:
        """Return a set of required packages to install in the build environment."""
        return set()

    def get_build_environment(self) -> dict[str, str]:
        """Return a dictionary with the environment to use in the build step."""
        return {
            # Cryptography fails to load OpenSSL legacy provider in some circumstances.
            # Since we don't need the legacy provider, this works around that bug.
            "CRYPTOGRAPHY_OPENSSL_NO_LEGACY": "true"
        }

    def get_build_commands(self) -> list[str]:
        """Return a list of commands to run during the build step."""
        options = cast(ReactivePluginProperties, self._options)

        command = [
            sys.executable,
            "-I",
            __file__,
            self._part_info.project_name,
            str(self._part_info.part_build_dir),
            str(self._part_info.part_install_dir),
        ]
        # The YAML List[str] schema would colocate any options with arguments
        # in the same string.  This is not what we want as we need to send
        # these separately when calling out to the command later.
        #
        # Expand any such strings as we add them to the command.
        for arg in options.reactive_charm_build_arguments:
            command.extend(shlex.split(arg))
        return [" ".join(shlex.quote(i) for i in command)]


def run_charm_tool(args: list[str]):
    """Run the charm tool, log and check exit code."""
    result_classification = "SUCCESS"
    exc = None

    print(f"charm tool execution command={args}")
    try:
        completed_process = subprocess.run(args, check=True)
    except subprocess.CalledProcessError as call_error:
        exc = call_error
        if call_error.returncode < 100 or call_error.returncode >= 200:
            result_classification = "ERROR"
            raise
        result_classification = "WARNING"
        print(f"charm tool execution {result_classification}: returncode={exc.returncode}")
    else:
        print(
            f"charm tool execution {result_classification}: returncode={completed_process.returncode}"
        )


def build(
    *, charm_name: str, build_dir: Path, install_dir: Path, charm_build_arguments: list[str]
):
    """Build a charm using charm tool.

    The charm tool is used to build reactive charms, the build process
    is as follows:

    - Run charm proof to ensure the charm
      would build with no errors (warnings are allowed)

    - Link charm tool's build directory to the part lifecycle's
      install_dir

    - Run "charm build"

    Note that no files/dirs in the original project are modified nor removed
    because in that case the VCS will detect something changed and the version
    string produced by `charm` would be misleading.
    """
    # Verify the charm is ok from a charm tool point of view.

    try:
        run_charm_tool(["charm", "proof"])
    except subprocess.CalledProcessError as call_error:
        return call_error.returncode

    # Link the installation directory to the place where charm creates
    # the charm.
    charm_build_dir = build_dir / charm_name
    if not charm_build_dir.exists():
        charm_build_dir.symlink_to(install_dir, target_is_directory=True)

    cmd = ["charm", "build"]
    if charm_build_arguments:
        cmd.extend(charm_build_arguments)
    cmd.extend(["-o", str(build_dir)])

    try:
        run_charm_tool(cmd)
    except subprocess.CalledProcessError as call_error:
        return call_error.returncode
    finally:
        charm_build_dir.unlink()

    return 0


if __name__ == "__main__":
    returncode = build(
        charm_name=sys.argv[1],
        build_dir=Path(sys.argv[2]),
        install_dir=Path(sys.argv[3]),
        charm_build_arguments=sys.argv[4:],
    )
    sys.exit(returncode)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import datetime

import pydantic
from craft_store import models


def get_fake_revision(**kwargs) -> models.resource_revision_model.CharmResourceRevision:
    data = {
        "bases": [],
        "created_at": datetime.datetime(2024, 1, 2, 4, 8, 16),
        "name": "my-resource",
        "revision": 0,
        "sha256": "sha256",
        "sha3_384": "sha3_384",
        "sha384": "sha384",
        "sha512": "sha512",
        "size": pydantic.ByteSize(0),
        "type": "file",
    }
    data.update(kwargs)
    return models.resource_revision_model.CharmResourceRevision(**data)



# This file is part of charmcraft.
#
# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import datetime
import os
import pathlib
import subprocess
import sys

import craft_parts_docs

import charmcraft

project_dir = pathlib.Path(__file__).parents[1].resolve()

project = "Charmcraft"
author = "Canonical"
release = charmcraft.__version__
if ".post" in release:
    # The commit hash in the dev release version confuses the spellchecker
    release = "dev"

copyright = "2023-%s, %s" % (datetime.date.today().year, author)

# region Configuration for canonical-sphinx
ogp_site_url = "https://canonical-charmcraft.readthedocs-hosted.com/"
ogp_site_name = project
ogp_image = "https://assets.ubuntu.com/v1/253da317-image-document-ubuntudocs.svg"

html_context = {
    "product_page": "github.com/canonical/charmcraft",
    "github_url": "https://github.com/canonical/charmcraft",
    "discourse": "https://discourse.charmhub.io",
}

extensions = [
    "canonical_sphinx",
]
# endregion

# region General configuration
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions.extend(
    [
        "sphinx.ext.intersphinx",
        "sphinx.ext.viewcode",
        "sphinx.ext.coverage",
        "sphinx.ext.doctest",
        "sphinx-pydantic",
        "sphinx_toolbox",
        "sphinx_toolbox.more_autodoc",
        "sphinx.ext.autodoc",  # Must be loaded after more_autodoc
        "sphinx_autodoc_typehints",  # must be loaded after napoleon
        "sphinxcontrib.details.directive",
        "sphinx_toolbox.collapse",
        "sphinxcontrib.autodoc_pydantic",
    ]
)

# endregion

autodoc_default_options = {"exclude-members": "model_post_init"}

# region Options for extensions
# Intersphinx extension
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#configuration

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "craft-parts": ("https://canonical-craft-parts.readthedocs-hosted.com/en/latest/", None),
}
# See also:
# https://www.sphinx-doc.org/en/master/usage/extensions/intersphinx.html#confval-intersphinx_disabled_reftypes
intersphinx_disabled_reftypes = ["*"]

# Type hints configuration
set_type_checking_flag = True
typehints_fully_qualified = False
always_document_param_types = True

# Github config
github_username = "canonical"
github_repository = "charmcraft"

html_domain_indices = True

# endregion

# Setup libraries documentation snippets for use in charmcraft docs.
# TODO: This needs a craft-parts update first.
# common_docs_path = pathlib.Path(__file__).parent / "common"
# craft_parts_docs_path = pathlib.Path(craft_parts_docs.__file__).parent / "craft-parts"
# (common_docs_path / "craft-parts").unlink(missing_ok=True)
# (common_docs_path / "craft-parts").symlink_to(
#     craft_parts_docs_path, target_is_directory=True
# )


def generate_cli_docs(nil):
    gen_cli_docs_path = (project_dir / "tools" / "gen_cli_docs.py").resolve()
    subprocess.run([sys.executable, gen_cli_docs_path, project_dir / "docs"], check=True)


def setup(app):
    app.connect("builder-inited", generate_cli_docs)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Store helper utilities."""
from collections.abc import Iterable

from craft_store import endpoints


def get_packages(
    charms: Iterable[str] = (), bundles: Iterable[str] = ()
) -> list[endpoints.Package]:
    """Get a list of packages from charms and bundles."""
    return [
        *(endpoints.Package(package_type="charm", package_name=charm) for charm in charms),
        *(endpoints.Package(package_type="bundle", package_name=bundle) for bundle in bundles),
    ]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Service class for running craft lifecycle commands."""
from __future__ import annotations

from craft_application import services, util
from craft_cli import emit
from overrides import override


class LifecycleService(services.LifecycleService):
    """Business logic for lifecycle builds."""

    def setup(self) -> None:
        """Do Charmcraft-specific setup work."""
        self._manager_kwargs.setdefault("project_name", self._project.name)
        super().setup()

    @override
    def _get_build_for(self) -> str:
        build_for = super()._get_build_for()
        if "-" not in build_for:
            if self._build_plan and self._build_plan[0].build_for == "all":
                emit.progress(
                    "WARNING: Charmcraft does not validate that charms with "
                    "architecture 'all' are fully architecture agnostic.",
                    permanent=True,
                )
            return build_for

        # Multi-arch builds: Tell craft-parts that we're building for any foreign
        # architecture (to trick it into trying to cross-compile anything, leading
        # to more likely failures.)
        emit.progress(
            "WARNING: Charmcraft does not validate that charms with multiple "
            "given architectures are architecture agnostic.",
            permanent=True,
        )
        host_arch = util.get_host_architecture()
        for arch in build_for.split("-"):
            if arch != host_arch:
                return arch

        return host_arch



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Platform-related Charmcraft utilities."""
import dataclasses
import pathlib
import platform
from collections.abc import Iterable

import distro
from craft_application import errors
from craft_parts.utils.formatting_utils import humanize_list

from charmcraft import const


@dataclasses.dataclass(frozen=True)
class OSPlatform:
    """Description of an operating system platform."""

    system: str
    release: str
    machine: str


def get_os_platform(filepath: pathlib.Path = pathlib.Path("/etc/os-release")) -> OSPlatform:
    """Determine a system/release combo for an OS using /etc/os-release if available."""
    system = platform.system()
    release = platform.release()
    machine = platform.machine()

    if system == "Linux":
        info = distro.info()
        system = info.get("id", system)
        # Treat Ubuntu derivatives as Ubuntu, as they should be compatible.
        if system != "ubuntu" and "ubuntu" in info.get("like", "").split():
            system = "ubuntu"
        release = info.get("version", release)

    return OSPlatform(system=system, release=release, machine=machine)


def validate_architectures(architectures: Iterable[str], *, allow_all: bool = False) -> None:
    """Validate that all architectures provided are valid architecture names."""
    architectures = set(architectures)
    if allow_all and "all" in architectures:
        if architectures == {"all"}:
            return
        raise errors.CraftValidationError(
            "If 'all' is defined for architectures, it must be the only architecture."
        )
    invalid = architectures - const.SUPPORTED_ARCHITECTURES
    if invalid:
        raise errors.CraftValidationError(
            f"Invalid architecture(s): {', '.join(invalid)}",
            details=f"Valid architecture values are: {humanize_list(sorted(const.SUPPORTED_ARCHITECTURES), 'and')}",
        )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Tests for metadata models."""
import json

import pytest

from charmcraft.models import metadata, project

BASIC_CHARM_METADATA_DICT = {
    "name": "test-charm",
    "summary": "A charm for testing, with a summary string that is more than seventy-eight characters long.",
    "description": "A fake charm used for testing purposes.",
}
BASIC_CHARM_DICT = {
    **BASIC_CHARM_METADATA_DICT,
    "type": "charm",
    "bases": [{"name": "ubuntu", "channel": "22.04", "architectures": ["riscv64"]}],
}

BASIC_BUNDLE_METADATA_DICT = {
    "name": "test-bundle",
    "description": "A fake bundle for testing purposes.",
}
BASIC_BUNDLE_DICT = {
    "type": "bundle",
    **BASIC_BUNDLE_METADATA_DICT,
}


@pytest.mark.parametrize(
    ("charm_dict", "expected"),
    [
        (BASIC_CHARM_DICT, BASIC_CHARM_METADATA_DICT),
        (
            dict(**BASIC_CHARM_DICT, links={"documentation": "https://docs.url/"}),
            {**BASIC_CHARM_METADATA_DICT, "docs": "https://docs.url/"},
        ),
        (
            dict(**BASIC_CHARM_DICT, links={"contact": "someone@company.com"}),
            {**BASIC_CHARM_METADATA_DICT, "maintainers": ["someone@company.com"]},
        ),
        (
            dict(**BASIC_CHARM_DICT, links={"contact": ["someone@company.com"]}),
            {**BASIC_CHARM_METADATA_DICT, "maintainers": ["someone@company.com"]},
        ),
        pytest.param(
            dict(
                **BASIC_CHARM_DICT,
                links={"issues": "https://github.com/canonical/charmcraft/issues"},
            ),
            {
                **BASIC_CHARM_METADATA_DICT,
                "issues": "https://github.com/canonical/charmcraft/issues",
            },
            id="non-transformed-link",
        ),
        (
            dict(**BASIC_CHARM_DICT, title="Title becomes display name"),
            {**BASIC_CHARM_METADATA_DICT, "display-name": "Title becomes display name"},
        ),
        pytest.param(
            {**BASIC_CHARM_DICT, "charm-user": "sudoer"},
            {**BASIC_CHARM_METADATA_DICT, "charm-user": "sudoer"},
            id="sudoer",
        ),
    ],
)
def test_charm_metadata_from_charm_success(charm_dict, expected):
    charm = project.CharmcraftProject.unmarshal(charm_dict)

    assert json.loads(json.dumps(metadata.CharmMetadata.from_charm(charm).marshal())) == expected


@pytest.mark.parametrize(
    ("bundle_dict", "expected"),
    [
        (BASIC_BUNDLE_DICT, BASIC_BUNDLE_METADATA_DICT),
    ],
)
def test_bundle_metadata_from_bundle(bundle_dict, expected):
    bundle = project.Bundle.unmarshal(BASIC_BUNDLE_DICT)

    assert (
        json.loads(json.dumps(metadata.BundleMetadata.from_bundle(bundle).marshal())) == expected
    )



# Copyright 2020-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Indicate which files are ignored by Juju.

This module is used by charm_builder.py script in a separate process than Charmcraft.
"""

import re
import typing

KEEP = "keep"
SKIP = "skip"
FORCEKEEP = "forcekeep"

_unescapes = {
    r"\!": "!",
    r"\ ": " ",
    r"\#": "#",
}


def _rstrip_unescaped(rule):
    """Remove trailing whitespace that isn't escaped."""
    i = len(rule) - 1
    last = len(rule)
    while i >= 0:
        if rule[i] == "\n" or rule[i] == "\r":
            last = i
        elif rule[i] != " ":
            break
        elif i == 0 or rule[i - 1] != "\\":
            last = i
        i -= 1
    return rule[:last]


def _unescape_rule(rule):
    """Take out escape characters and trailing unescaped whitespace from the rule."""
    rule = rule.lstrip()
    rule = _rstrip_unescaped(rule)
    for old, new in _unescapes.items():
        rule = rule.replace(old, new)
    return rule


def _rule_to_regex(rule):
    """Turn a rule into a regex that we can use.

    This assumes that all the meta processing, like 'ends with /' and 'starts with !' have
    already been checked.
    """
    # Things we currently care about:
    # * = matches only within a directory "[^/]*"
    # ** = matches across directories ".*"
    # ? = matches a single character
    # [0-9] can match anything 0-9
    # This is taken from fnmatch.fnmatch, but that doesn't handle '**' and '*' also matches
    # directories

    i, n = 0, len(rule)
    res = ""
    while i < n:
        c = rule[i]
        i += 1
        if c == "*":
            if i < n and rule[i] == "*":
                i += 1
                res += ".*"
            else:
                res += "[^/]*"
        elif c == "?":
            res += "[^/]"
        elif c == "[":
            j = i
            if j < n and rule[j] == "!":
                j += 1
            if j < n and rule[j] == "]":
                j += 1
            while j < n and rule[j] != "]":
                j += 1
            if j >= n:
                res += "\\["
            else:
                stuff = rule[i:j]
                # Escape regex set operations (&~|).
                stuff = re.sub(r"([&~|])", r"\\\1", stuff)
                i = j + 1
                if stuff[0] == "!":
                    stuff = "^" + stuff[1:]
                elif stuff[0] in ("["):
                    stuff = "\\" + stuff
                res = f"{res}[{stuff}]"
        elif c == "/":
            # Special case of '/**/' which can match a single '/'
            if i < n and rule[i] == "*" and rule[i - 1 : i + 3] == "/**/":
                i += 3
                res = res + ".*/"
            else:
                res = res + "/"
        else:
            res += re.escape(c)
    res += r"\Z"
    return res


class _Matcher:
    """Couple a regex with other metadata for how we should match a given pattern."""

    def __init__(
        self,
        line_num: int,
        orig_rule: str,
        invert: bool,
        only_dirs: bool,
        regex: typing.Pattern,
    ):
        self.line_num = line_num
        self.orig_rule = orig_rule
        self.invert = invert
        self.only_dirs = only_dirs
        self.compiled = re.compile(regex, re.DOTALL)

    def match(self, path: str, is_dir: bool) -> str:
        """Check if a path matches.

        Returns
        -------
            Can return one of KEEP, SKIP, FORCEKEEP

        """
        if self.only_dirs and not is_dir:
            return KEEP
        if self.compiled.match(path):
            if self.invert:
                return FORCEKEEP
            return SKIP
        return KEEP


class JujuIgnore:
    """Track a set of ignore patterns from a .jujuignore file."""

    def __init__(self, patterns: typing.Iterable[str]):
        self._matchers: list[_Matcher] = []
        self._compile_from(patterns)

    def extend_patterns(self, patterns: typing.Iterable[str]) -> None:
        """Add more patterns to the ignore list."""
        self._compile_from(patterns)

    def _compile_from(self, patterns: typing.Iterable[str]):
        for line_num, rule in enumerate(patterns, 1):
            orig_rule = rule
            rule = rule.lstrip().rstrip("\r\n")
            if not rule or rule.startswith("#"):
                continue
            invert = False
            if rule.startswith("!"):
                invert = True
                rule = rule.lstrip("!")
            rule = _unescape_rule(rule)
            only_dirs = False
            if rule.endswith("/"):
                only_dirs = True
                rule = rule.rstrip("/")
            if not rule.startswith("/"):
                # A rule that doesn't start with '/' means to match any
                # subdirectory
                rule = "**/" + rule
            regex = _rule_to_regex(rule)
            m = _Matcher(
                line_num=line_num,
                orig_rule=orig_rule,
                invert=invert,
                only_dirs=only_dirs,
                regex=regex,
            )
            self._matchers.append(m)
            print(f"Translated .jujuignore {line_num:d} {orig_rule!r} => {regex!r}")

    def match(self, path: str, is_dir: bool) -> bool:
        """Check if the given path should be ignored.

        :param path: A local path (eg /foo/bar or foo/bar) from the root directory of the project.
        :param is_dir: Indicate whether the given path is a directory (because of special handling
            from ignore files when the path ends with a '/')

        :returns: A boolean indicating whether the ignore rules matched the given path (thus the path
            should be ignored).

        """
        if not path.startswith("/"):
            path = "/" + path
        keep = True
        for matcher in self._matchers:
            matchRes = matcher.match(path, is_dir)
            if matchRes == SKIP:
                keep = False
            elif matchRes == FORCEKEEP:
                keep = True
                break
        return not keep


# default_juju_ignore is the initial set of ignores.
# juju itself always includes these before adding the contents of .jujuignore
# NOTE that this diverges from Juju ignore list, which also ignores "version",
# because we need the version file to populate the store
default_juju_ignore = """
.git
.svn
.hg
.bzr
.tox

/build/
/revision
/venv

.jujuignore
""".split(
    "\n"
)



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for skopeo wrapper."""

import contextlib
import os
import pathlib
import platform
import shutil

import pytest

from charmcraft import errors
from charmcraft.utils.skopeo import Skopeo

pytestmark = [
    pytest.mark.skipif(
        "CI" not in os.environ and not shutil.which("skopeo"), reason="skopeo not found in PATH"
    ),
    pytest.mark.xfail(
        platform.system().lower() not in ("linux", "darwin"),
        reason="Don't necessarily have skopeo on non Linux/mac platforms.",
        strict=False,  # Allow them to pass anyway.
    ),
]


@pytest.mark.parametrize(
    ("name", "image", "tag"),
    [
        ("alpine", "docker://ghcr.io/containerd/alpine", "3.14.0"),
        ("debian12", "docker://gcr.io/distroless/base-debian12", "nonroot"),
        ("mock-rock", "docker://ghcr.io/canonical/oci-factory/mock-rock", "1.2-22.04_279"),
        ("nanoserver", "docker://ghcr.io/containerd/nanoserver", "1809"),
    ],
)
def test_inspect_and_download(
    monkeypatch, tmp_path: pathlib.Path, name: str, image: str, tag: str
):
    (tmp_path / "tmp").mkdir()
    skopeo = Skopeo(tmpdir=tmp_path / "tmp")

    raw_data = skopeo.inspect(f"{image}:{tag}", raw=True)

    with contextlib.suppress(errors.SubprocessError):
        # These fail if the host platform doesn't match a valid platform for the container.
        meta_data = skopeo.inspect(f"{image}:{tag}")
        meta_data_by_hash = skopeo.inspect(f"{image}@{meta_data['Digest']}")
        assert meta_data == meta_data_by_hash
        raw_data_by_hash = skopeo.inspect(f"{image}@{meta_data['Digest']}", raw=True)
        assert raw_data == raw_data_by_hash

    for manifest in raw_data["manifests"]:
        if "variant" in manifest["platform"]:
            # We don't handle variants currently.
            continue
        os = manifest["platform"]["os"]
        arch = manifest["platform"]["architecture"]
        digest = manifest["digest"]
        by_digest_tar = tmp_path / f"{name}_{os}_{arch}-digest.tar"
        by_os_arch_tar = tmp_path / f"{name}_{os}_{arch}.tar"
        os_arch_skopeo = Skopeo(arch=arch, os=os)

        os_arch_skopeo.copy(f"{image}:{tag}", f"oci-archive:{by_os_arch_tar}")
        skopeo.copy(f"{image}@{digest}", f"oci-archive:{by_digest_tar}")

        assert by_digest_tar.exists()
        assert by_os_arch_tar.exists()

        image_info_by_digest = skopeo.inspect(f"{image}@{digest}")
        image_info_by_os_arch = os_arch_skopeo.inspect(f"{image}:{tag}")

        for key in ["Architecture", "Os", "Layers", "LayersData", "Env"]:
            assert image_info_by_digest[key] == image_info_by_os_arch[key]



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for store client."""

from unittest import mock

import pytest

from charmcraft import store


@pytest.fixture
def client() -> store.Client:
    return store.Client(api_base_url="http://charmhub.local")


@pytest.fixture
def anonymous_client() -> store.AnonymousClient:
    return store.AnonymousClient("http://charmhub.local", "http://storage.charmhub.local")


@pytest.mark.parametrize(
    ("charm", "lib_id", "api", "patch", "expected_call"),
    [
        (
            "my-charm",
            "abcdefg",
            None,
            None,
            mock.call("GET", "/v1/charm/libraries/my-charm/abcdefg", params={}),
        ),
        (
            "my-charm",
            "abcdefg",
            0,
            0,
            mock.call(
                "GET", "/v1/charm/libraries/my-charm/abcdefg", params={"api": 0, "patch": 0}
            ),
        ),
    ],
)
def test_get_library_success(
    monkeypatch, anonymous_client, charm, lib_id, api, patch, expected_call
):
    mock_get_urlpath_json = mock.Mock(
        return_value={
            "charm-name": charm,
            "library-name": "my_lib",
            "library-id": lib_id,
            "api": api or 1,
            "patch": patch or 2,
            "hash": "hashy!",
        }
    )
    monkeypatch.setattr(anonymous_client, "request_urlpath_json", mock_get_urlpath_json)

    anonymous_client.get_library(charm_name=charm, library_id=lib_id, api=api, patch=patch)

    mock_get_urlpath_json.assert_has_calls([expected_call])


@pytest.mark.parametrize(
    ("libs", "json_response", "expected"),
    [
        ([], {"libraries": []}, []),
        (
            [{"charm-name": "my-charm", "library-name": "my_lib"}],
            {
                "libraries": [
                    {
                        "charm-name": "my-charm",
                        "library-name": "my_lib",
                        "library-id": "ididid",
                        "api": 1,
                        "patch": 2,
                        "hash": "hashhashhash",
                    },
                ],
            },
            [
                store.models.Library(
                    charm_name="my-charm",
                    lib_name="my_lib",
                    lib_id="ididid",
                    api=1,
                    patch=2,
                    content=None,
                    content_hash="hashhashhash",
                ),
            ],
        ),
    ],
)
def test_fetch_libraries_metadata(monkeypatch, anonymous_client, libs, json_response, expected):
    mock_get_urlpath_json = mock.Mock(return_value=json_response)
    monkeypatch.setattr(anonymous_client, "request_urlpath_json", mock_get_urlpath_json)

    assert anonymous_client.fetch_libraries_metadata(libs) == expected

    mock_get_urlpath_json.assert_has_calls(
        [mock.call("POST", "/v1/charm/libraries/bulk", json=libs)]
    )



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for the lifecycle service."""

import pytest
from craft_application import models, util
from craft_cli.pytest_plugin import RecordingEmitter
from craft_providers import bases

from charmcraft.services.lifecycle import LifecycleService

HOST_ARCH = util.get_host_architecture()


@pytest.fixture
def service(service_factory) -> LifecycleService:
    return service_factory.lifecycle


@pytest.mark.parametrize(
    ("plan_build_for", "expected"),
    [
        ("amd64", "amd64"),
        ("arm64", "arm64"),
        ("riscv64", "riscv64"),
        ("all", HOST_ARCH),
        # Test multi-arch charms
        (f"{HOST_ARCH}-foreign", "foreign"),
        (f"foreign-{HOST_ARCH}", "foreign"),
    ],
)
def test_get_build_for_values(service: LifecycleService, plan_build_for: str, expected: str):
    service._build_plan = [
        models.BuildInfo(
            base=bases.BaseName("ubuntu", "22.04"),
            platform="something",
            build_on=HOST_ARCH,
            build_for=plan_build_for,
        )
    ]

    assert service._get_build_for() == expected


def test_build_for_warns_on_all(service: LifecycleService, emitter: RecordingEmitter):
    service._build_plan[0].build_for = "all"

    assert service._get_build_for() == HOST_ARCH

    emitter.assert_progress(
        "WARNING: Charmcraft does not validate that charms with architecture 'all' "
        "are fully architecture agnostic.",
        permanent=True,
    )


def test_build_for_warns_on_multi(service: LifecycleService, emitter: RecordingEmitter):
    service._build_plan[0].build_for = f"{HOST_ARCH}-foreign"

    assert service._get_build_for() == "foreign"

    emitter.assert_progress(
        "WARNING: Charmcraft does not validate that charms with multiple "
        "given architectures are architecture agnostic.",
        permanent=True,
    )



# Copyright 2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Integration tests for set-resource-architectures command."""
import argparse
import textwrap

import pytest
from craft_store import models

from charmcraft import application
from charmcraft.application.commands import SetResourceArchitecturesCommand
from tests import get_fake_revision


@pytest.fixture
def cmd(service_factory):
    return SetResourceArchitecturesCommand(
        config={"app": application.APP_METADATA, "services": service_factory}
    )


@pytest.mark.parametrize(
    ("args", "list_response", "expected_output"),
    [
        (
            ["my-charm", "my-resource", "--revision=1", "amd64,arm64"],
            [],
            "No revisions updated.",
        ),
        (
            ["--format=json", "my-charm", "my-resource", "--revision=1", "amd64,arm64"],
            [],
            "[]",
        ),
        (
            ["my-charm", "my-resource", "--revision=1", "amd64,arm64"],
            [
                get_fake_revision(
                    revision=1,
                    bases=[models.ResponseCharmResourceBase(architectures=["amd64", "arm64"])],
                ),
                get_fake_revision(
                    revision=2,
                    bases=[models.ResponseCharmResourceBase(architectures=["riscv64"])],
                ),
            ],
            textwrap.dedent(
                """\
                  Revision  Updated At    Architectures
                ----------  ------------  ---------------
                         1  --            amd64,arm64"""
            ),
        ),
        (
            ["my-charm", "my-resource", "--revision=1", "amd64,arm64", "--format=json"],
            [
                get_fake_revision(
                    revision=1,
                    bases=[models.ResponseCharmResourceBase(architectures=["amd64", "arm64"])],
                ),
                get_fake_revision(
                    revision=2,
                    bases=[models.ResponseCharmResourceBase(architectures=["riscv64"])],
                ),
            ],
            textwrap.dedent(
                """\
                [
                    {
                        "revision": 1,
                        "updated_at": null,
                        "architectures": [
                            "amd64",
                            "arm64"
                        ]
                    }
                ]"""
            ),
        ),
    ],
)
def test_set_resource_architectures(
    emitter, service_factory, cmd, args, list_response, expected_output
):
    """Test the happy path for set-resource-architectures command."""
    service_factory.store.client.list_resource_revisions.return_value = list_response

    parser = argparse.ArgumentParser()
    cmd.fill_parser(parser)

    parsed_args = parser.parse_args(args)

    cmd.run(parsed_args)

    emitter.assert_message(expected_output)



# Copyright 2020-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import contextlib
import importlib
import json
import os
import pathlib
import tempfile
import types
from collections.abc import Iterator
from unittest import mock

import craft_parts
import pytest
import responses as responses_module
import yaml
from craft_application import models, util
from craft_parts import callbacks, plugins
from craft_providers import bases

import charmcraft.parts
from charmcraft import const, env, instrum, parts, services, store
from charmcraft.application.main import APP_METADATA
from charmcraft.models import project


@pytest.fixture
def simple_charm():
    return project.BasesCharm(
        type="charm",
        name="charmy-mccharmface",
        summary="Charmy!",
        description="Very charming!",
        bases=[
            {
                "build-on": [
                    {
                        "name": "ubuntu",
                        "channel": "22.04",
                        "architectures": [util.get_host_architecture()],
                    }
                ],
                "run-on": [{"name": "ubuntu", "channel": "22.04", "architectures": ["arm64"]}],
            }
        ],
    )


@pytest.fixture
def mock_store_client():
    client = mock.Mock(spec_set=store.Client)

    client.whoami.return_value = {
        "account": {"username": "test-user"},
    }

    return client


@pytest.fixture
def mock_store_anonymous_client() -> mock.Mock:
    return mock.Mock(spec_set=store.AnonymousClient)


@pytest.fixture
def service_factory(
    fs,
    fake_project_dir,
    fake_prime_dir,
    simple_charm,
    mock_store_client,
    mock_store_anonymous_client,
    default_build_plan,
) -> services.CharmcraftServiceFactory:
    factory = services.CharmcraftServiceFactory(app=APP_METADATA)

    factory.set_kwargs(
        "package",
        project_dir=fake_project_dir,
        build_plan=default_build_plan,
    )
    factory.set_kwargs(
        "lifecycle",
        work_dir=pathlib.Path("/project"),
        cache_dir=pathlib.Path("/cache"),
        build_plan=default_build_plan,
    )

    factory.project = simple_charm

    factory.store.client = mock_store_client
    factory.store.anonymous_client = mock_store_anonymous_client

    return factory


@pytest.fixture
def default_build_plan():
    arch = util.get_host_architecture()
    return [
        models.BuildInfo(
            base=bases.BaseName("ubuntu", "22.04"),
            build_on=arch,
            build_for="arm64",
            platform="distro-1-test64",
        )
    ]


@pytest.fixture
def fake_project_dir(fs) -> pathlib.Path:
    project_dir = pathlib.Path("/root/project")
    fs.create_dir(project_dir)
    return project_dir


@pytest.fixture
def fake_prime_dir(fs) -> pathlib.Path:
    prime_dir = pathlib.Path("/root/prime")
    fs.create_dir(prime_dir)
    return prime_dir


@pytest.fixture
def fake_path(fs) -> Iterator[pathlib.Path]:
    """Like tmp_path, but with a fake filesystem."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        yield pathlib.Path(tmp_dir)


@pytest.fixture
def global_debug():
    os.environ["CRAFT_DEBUG"] = "1"


@pytest.fixture
def new_path(monkeypatch, tmp_path):
    monkeypatch.chdir(tmp_path)
    return tmp_path


@pytest.fixture(autouse=True, scope="session")
def tmpdir_under_tmpdir(tmpdir_factory):
    tempfile.tempdir = str(tmpdir_factory.getbasetemp())


@pytest.fixture(autouse=True, scope="session")
def setup_parts():
    parts.setup_parts()


@pytest.fixture
def charmhub_config() -> env.CharmhubConfig:
    """Provide a charmhub config for use in tests"""
    return env.CharmhubConfig(
        api_url="https://api.staging.charmhub.io",
        storage_url="https://storage.staging.snapcraftcontent.com",
        registry_url="https://registry.staging.jujucharms.com",
    )


@pytest.fixture(autouse=True)
def intertests_cleanups():
    """Run some cleanups between tests.

    Before each test:

    - reload the instrumentator module to start clean.

    After each test:

    - clear the already-notified structure for each test (this is needed as that
      structure is a module-level one (by design), so otherwise it will be dirty
      between tests).

    - unregister all Craft Parts plugins callbacks
    """
    importlib.reload(instrum)
    yield
    callbacks.unregister_all()


@pytest.fixture
def responses():
    """Simple helper to use responses module as a fixture, for easier integration in tests."""
    with responses_module.RequestsMock() as rsps:
        yield rsps


@pytest.fixture
def prepare_charmcraft_yaml(tmp_path: pathlib.Path):
    """Helper to create a charmcraft.yaml file in disk.

    If content is not given, remove charmcraft.yaml if exists.
    """

    charmcraft_yaml_path = tmp_path / const.CHARMCRAFT_FILENAME

    def prepare_charmcraft_yaml(content: str | None = None):
        if content is None:
            with contextlib.suppress(OSError):
                charmcraft_yaml_path.unlink(missing_ok=True)
        else:
            charmcraft_yaml_path.write_text(content)

        return tmp_path

    return prepare_charmcraft_yaml


def prepare_file(tmp_path: pathlib.Path, filename: str):
    """Helper to create a file under a temporary path."""

    path = tmp_path / filename

    def prepare(content: str | None = None):
        if content is None:
            path.unlink(missing_ok=True)
        else:
            path.write_text(content)

        return tmp_path

    return prepare


@pytest.fixture
def prepare_metadata_yaml(tmp_path: pathlib.Path):
    """Helper to create a metadata.yaml file in disk.

    If content is not given, remove metadata.yaml if exists.
    """
    return prepare_file(tmp_path, const.METADATA_FILENAME)


@pytest.fixture
def prepare_actions_yaml(tmp_path: pathlib.Path):
    """Helper to create a actions.yaml file in disk.

    If content is not given, remove actions.yaml if exists.
    """
    return prepare_file(tmp_path, const.JUJU_ACTIONS_FILENAME)


@pytest.fixture
def prepare_config_yaml(tmp_path: pathlib.Path):
    """Helper to create a config.yaml file in disk.

    If content is not given, remove config.yaml if exists.
    """
    return prepare_file(tmp_path, const.JUJU_CONFIG_FILENAME)


@pytest.fixture
def emitter(emitter):
    """Monkeypatch craft-cli's emitter fixture to easily test the JSON encoded output."""

    def assert_json_output(self, expected_content):
        """Get last output, which should be a message, and validate its content."""
        last_output = self.interactions[-1]
        output_type, raw_output = last_output.args
        assert output_type == "message", "Last command output is not 'message'"
        try:
            output_content = json.loads(raw_output)
        except json.decoder.JSONDecodeError:
            pytest.fail("Last command output is not valid JSON.")
        assert output_content == expected_content

    emitter.assert_json_output = types.MethodType(assert_json_output, emitter)
    return emitter


@pytest.fixture
def assert_output(capsys):
    """Assert that a given string was sent to stdout.

    This is a helper to simplify tests for charm_builder.py and its modules that print
    directly to stdout.

    Note that every call to this helper will clear the previous captured output.
    """

    def helper(*match_lines):
        captured = capsys.readouterr()
        printed_lines = captured.out.splitlines()
        for match_line in match_lines:
            if match_line not in printed_lines:
                printed_repr = "\n".join(map(repr, printed_lines))
                pytest.fail(f"Line {match_line!r} not found in the output found:\n{printed_repr}")

    return helper


@pytest.fixture
def build_charm_directory():
    def helper(tmp_path, fake_charms, file_type="charm"):
        expected = {}
        charmcraft_yaml = {"type": file_type}
        for name, path in fake_charms.items():
            full_path = tmp_path / path
            expected[name] = full_path
            full_path.mkdir(parents=True)
            metadata_yaml = {"name": name}
            with (full_path / const.CHARMCRAFT_FILENAME).open("w") as yaml_file:
                yaml.safe_dump(charmcraft_yaml, yaml_file)
            with (full_path / const.METADATA_FILENAME).open("w") as yaml_file:
                yaml.safe_dump(metadata_yaml, yaml_file)
        return expected

    return helper


@pytest.fixture
def stub_extensions(monkeypatch):
    from charmcraft.extensions import registry

    extensions_dict = {}
    monkeypatch.setattr(registry, "_EXTENSIONS", extensions_dict)

    return extensions_dict


@pytest.fixture
def charm_plugin(tmp_path):
    requirement_files = ["reqs1.txt", "reqs2.txt"]
    for req in requirement_files:
        (tmp_path / req).write_text("somedep")
    project_dirs = craft_parts.ProjectDirs(work_dir=tmp_path)
    spec = {
        "plugin": "charm",
        "source": str(tmp_path),
        "charm-entrypoint": "entrypoint",
        "charm-binary-python-packages": ["pkg1", "pkg2"],
        "charm-python-packages": ["pkg3", "pkg4"],
        "charm-requirements": requirement_files,
    }
    plugin_properties = parts.CharmPluginProperties.unmarshal(spec)
    part_spec = plugins.extract_part_properties(spec, plugin_name="charm")
    part = craft_parts.Part(
        "foo", part_spec, project_dirs=project_dirs, plugin_properties=plugin_properties
    )
    project_info = craft_parts.ProjectInfo(
        application_name="test",
        project_dirs=project_dirs,
        cache_dir=tmp_path,
    )
    part_info = craft_parts.PartInfo(project_info=project_info, part=part)

    return plugins.get_plugin(part=part, part_info=part_info, properties=plugin_properties)


@pytest.fixture
def bundle_plugin(tmp_path):
    project_dirs = craft_parts.ProjectDirs(work_dir=tmp_path)
    spec = {
        "plugin": "bundle",
        "source": str(tmp_path),
    }
    plugin_properties = charmcraft.parts.bundle.BundlePluginProperties.unmarshal(spec)
    part_spec = plugins.extract_part_properties(spec, plugin_name="bundle")
    part = craft_parts.Part(
        "foo", part_spec, project_dirs=project_dirs, plugin_properties=plugin_properties
    )
    project_info = craft_parts.ProjectInfo(
        application_name="test",
        project_dirs=project_dirs,
        cache_dir=tmp_path,
    )
    part_info = craft_parts.PartInfo(project_info=project_info, part=part)

    return plugins.get_plugin(part=part, part_info=part_info, properties=plugin_properties)



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft environment utilities."""
import dataclasses
import os
import pathlib

import platformdirs
from craft_application.util import strtobool

from charmcraft import const


def get_host_shared_cache_path() -> pathlib.Path:
    """Path for host shared cache."""
    shared_cache_env = os.getenv(const.SHARED_CACHE_ENV_VAR)
    if shared_cache_env is not None:
        cache_path = pathlib.Path(shared_cache_env).expanduser().resolve()
        cache_path.mkdir(parents=True, exist_ok=True)
        return cache_path

    return platformdirs.user_cache_path(appname="charmcraft", ensure_exists=True)


def get_managed_environment_home_path() -> pathlib.Path:
    """Path for home when running in managed environment."""
    return pathlib.Path("/root")


def get_managed_environment_log_path() -> pathlib.Path:
    """Path for charmcraft log when running in managed environment."""
    return pathlib.Path("/tmp/charmcraft.log")


def get_charm_builder_metrics_path() -> pathlib.Path:
    """Path for charmcraft metrics when running charm_builder."""
    return pathlib.Path("/tmp/charm_builder_metrics.json")


def get_managed_environment_project_path() -> pathlib.Path:
    """Path for project when running in managed environment."""
    return get_managed_environment_home_path() / "project"


def is_charmcraft_running_from_snap() -> bool:
    """Check if charmcraft is running from the snap."""
    return os.getenv("SNAP_NAME") == "charmcraft" and os.getenv("SNAP") is not None


def is_charmcraft_running_in_managed_mode() -> bool:
    """Check if charmcraft is running in a managed environment."""
    managed_flag = os.getenv(const.MANAGED_MODE_ENV_VAR, os.getenv("CRAFT_MANAGED_MODE", "n"))
    return strtobool(managed_flag)


@dataclasses.dataclass(frozen=True)
class CharmhubConfig:
    """Definition of Charmhub endpoint configuration."""

    api_url: str = "https://api.charmhub.io"
    storage_url: str = "https://storage.snapcraftcontent.com"
    registry_url: str = "https://registry.jujucharms.com"


DEFAULT_CHARMHUB_CONFIG = CharmhubConfig()


def get_store_config() -> CharmhubConfig:
    """Get the appropriate configuration for the store."""
    api_url = os.getenv(const.STORE_API_ENV_VAR, DEFAULT_CHARMHUB_CONFIG.api_url)
    storage_url = os.getenv(const.STORE_STORAGE_ENV_VAR, DEFAULT_CHARMHUB_CONFIG.storage_url)
    registry_url = os.getenv(const.STORE_REGISTRY_ENV_VAR, DEFAULT_CHARMHUB_CONFIG.registry_url)
    return CharmhubConfig(api_url=api_url, storage_url=storage_url, registry_url=registry_url)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Unit tests for analysis service."""
import pathlib
import tempfile
import zipfile
from unittest import mock

import pytest
import pytest_check

from charmcraft import application, linters
from charmcraft.models.lint import CheckResult, CheckType, LintResult
from charmcraft.services import analysis


class StubAttributeChecker(linters.AttributeChecker):
    def __init__(self, name, url, text, result):
        self.name = name
        self.url = url
        self.text = text
        self.result = result

    def __call__(self):
        return self

    def run(self, basedir: pathlib.Path) -> str:
        return self.result


class StubLinter(linters.Linter):
    def __init__(self, name, result):
        self.name = name
        self.url = f"http://example.org/{name}"
        self.text = f"returns {name}"
        self.result = result

    def __call__(self):
        return self

    def run(self, basedir: pathlib.Path) -> str:
        return self.result


STUB_ATTRIBUTE_CHECKERS = [
    StubAttributeChecker(
        "unknown_attribute", "https://example.com/unknown", "returns unknown", LintResult.UNKNOWN
    ),
    StubAttributeChecker("says_python", "https://python.org", "returns python", "python"),
]
STUB_CHECKER_RESULTS = [
    CheckResult(linter.name, linter.result, linter.url, CheckType.ATTRIBUTE, linter.text)
    for linter in STUB_ATTRIBUTE_CHECKERS
]
ATTRIBUTE_CHECKER_NAMES = frozenset(checker.name for checker in STUB_ATTRIBUTE_CHECKERS)
STUB_LINTERS = [
    StubLinter("success", LintResult.OK),
    StubLinter("warn", LintResult.WARNING),
    StubLinter("error", LintResult.ERROR),
    StubLinter("fatal", LintResult.FATAL),
    StubLinter("ignore", LintResult.IGNORED),
    StubLinter("???", LintResult.UNKNOWN),
    StubLinter("N/A", LintResult.NONAPPLICABLE),
]
STUB_LINTER_RESULTS = [
    CheckResult(linter.name, linter.result, linter.url, CheckType.LINT, linter.text)
    for linter in STUB_LINTERS
]
LINTER_NAMES = frozenset(linter.name for linter in STUB_LINTERS)
ALL_CHECKER_NAMES = ATTRIBUTE_CHECKER_NAMES | LINTER_NAMES


@pytest.fixture
def mock_temp_dir(monkeypatch):
    mock_obj = mock.MagicMock(spec=tempfile.TemporaryDirectory)
    monkeypatch.setattr(tempfile, "TemporaryDirectory", mock.Mock(return_value=mock_obj))
    return mock_obj


@pytest.fixture
def mock_zip_file(monkeypatch):
    mock_obj = mock.MagicMock(spec=zipfile.ZipFile)
    monkeypatch.setattr(zipfile, "ZipFile", mock.Mock(return_value=mock_obj))
    return mock_obj


@pytest.fixture
def analysis_service():
    return analysis.AnalysisService(app=application.APP_METADATA, services=None)


@pytest.mark.parametrize(
    ("checkers", "expected"),
    [(STUB_ATTRIBUTE_CHECKERS, STUB_CHECKER_RESULTS), (STUB_LINTERS, STUB_LINTER_RESULTS)],
)
def test_lint_directory_results(monkeypatch, analysis_service, checkers, expected):
    monkeypatch.setattr(linters, "CHECKERS", checkers)

    assert list(analysis_service.lint_directory(pathlib.Path())) == expected


@pytest.mark.parametrize("checkers", [STUB_ATTRIBUTE_CHECKERS + STUB_LINTERS])
@pytest.mark.parametrize(
    "ignore", [set(), {"success"}, ATTRIBUTE_CHECKER_NAMES, LINTER_NAMES, ALL_CHECKER_NAMES]
)
def test_lint_directory_ignores(monkeypatch, analysis_service, checkers, ignore):
    monkeypatch.setattr(linters, "CHECKERS", checkers)
    checker_names = {checker.name for checker in checkers}

    results = list(
        analysis_service.lint_directory(pathlib.Path(), ignore=ignore, include_ignored=False)
    )
    checkers_run = {r.name for r in results}

    pytest_check.is_true(checkers_run.isdisjoint(ignore), f"{checkers_run & ignore}")
    pytest_check.is_true(checkers_run.issubset(checker_names), str(checkers_run - checker_names))


def test_lint_file_results(fs, mock_temp_dir, mock_zip_file, monkeypatch, analysis_service):
    fake_charm = pathlib.Path("/fake/charm.charm")
    fs.create_file(fake_charm)
    mock_checker = mock.Mock()
    monkeypatch.setattr(linters, "CHECKERS", [mock.Mock(return_value=mock_checker)])
    fake_temp_path = pathlib.Path(mock_temp_dir.__enter__.return_value)

    results = list(analysis_service.lint_file(fake_charm))

    with pytest_check.check:
        mock_zip_file.__enter__.return_value.extractall.assert_called_once_with(fake_temp_path)
    with pytest_check.check:
        mock_checker.get_result.assert_called_once_with(fake_temp_path)
    pytest_check.equal(results, [mock_checker.get_result.return_value])



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""Models for linters."""
import enum
from typing import final

from pydantic import dataclasses


# Making this a str subclass makes it JSON serialisable as that string.
# This can be replaced with StrEnum once we can drop support for Python < 3.11
class CheckType(str, enum.Enum):
    """Type of analyzer, either attribute check or linter.

    More documentation: https://juju.is/docs/sdk/charmcraft-analyzers-and-linters.
    """

    ATTRIBUTE = "attribute"
    LINT = "lint"


@final
class LintResult:
    """Check results."""

    OK = "ok"
    WARNING = "warning"
    ERROR = "error"
    FATAL = "fatal"
    IGNORED = "ignored"
    UNKNOWN = "unknown"
    NONAPPLICABLE = "nonapplicable"


class ResultLevel(enum.IntEnum):
    """The level of a lint result."""

    UNKNOWN = -1
    OK = 0
    IGNORED = 1
    WARNING = 2
    ERROR = 3
    FATAL = 4

    @classmethod
    def from_result(cls, result: str) -> "ResultLevel":
        """Convert a linting result string to a ResultLevel."""
        try:
            return cls[result.upper()]
        except KeyError:
            return cls.UNKNOWN

    @property
    def return_code(self) -> int:
        """Get an application return code for this lint level."""
        if self == self.FATAL:
            return 1
        if self == self.ERROR:
            return 2
        if self == self.WARNING:
            return 3
        return 0


@dataclasses.dataclass(frozen=True)
class CheckResult:
    """The result of a single linter check."""

    name: str
    result: str
    url: str
    check_type: CheckType
    text: str

    @property
    def level(self) -> ResultLevel:
        """Get the error level of the result."""
        return ResultLevel.from_result(self.result)

    def __str__(self) -> str:
        if self.result == LintResult.IGNORED or not self.result:
            info = ""
        elif self.result == LintResult.UNKNOWN:
            info = self.text
        else:
            info = f"[{self.result.upper()}] {self.text}".strip()

        if info:
            return f"{self.name}: {info} ({self.url})"
        return f"{self.name}: ({self.url}) {info}"



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import sys

import pytest

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows not supported")


def test_bundleplugin_get_build_package(bundle_plugin):
    assert bundle_plugin.get_build_packages() == set()


def test_bundleplugin_get_build_snaps(bundle_plugin):
    assert bundle_plugin.get_build_snaps() == set()


def test_bundleplugin_get_build_environment(bundle_plugin):
    assert bundle_plugin.get_build_environment() == {}


def test_bundleplugin_get_build_commands(bundle_plugin, tmp_path):
    if sys.platform == "linux":
        assert bundle_plugin.get_build_commands() == [
            f'mkdir -p "{str(tmp_path)}/parts/foo/install"',
            f'cp --archive --link --no-dereference {tmp_path}/parts/foo/build/* "{str(tmp_path)}/parts/foo/install"',
        ]
    else:
        assert bundle_plugin.get_build_commands() == [
            f'mkdir -p "{str(tmp_path)}/parts/foo/install"',
            f'cp -R -p -P {tmp_path}/parts/foo/build/* "{str(tmp_path)}/parts/foo/install"',
        ]



#!/usr/bin/env python3
# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import sys
import yaml

input_yaml = yaml.safe_load(sys.stdin)
with open(sys.argv[1], "r", encoding="UTF-8") as f:
    compare_yaml = yaml.safe_load(f)

if input_yaml != compare_yaml:
    sys.exit(1)



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Constants used in charmcraft."""
import enum
from typing import Literal

from craft_providers.bases import BaseName

# region Environment variables
ALTERNATE_AUTH_ENV_VAR = "CHARMCRAFT_AUTH"
DEVELOPER_MODE_ENV_VAR = "CHARMCRAFT_DEVELOPER"
EXPERIMENTAL_EXTENSIONS_ENV_VAR = "CHARMCRAFT_ENABLE_EXPERIMENTAL_EXTENSIONS"
IMAGE_INFO_ENV_VAR = "CHARMCRAFT_IMAGE_INFO"
PROVIDER_ENV_VAR = "CHARMCRAFT_PROVIDER"
SHARED_CACHE_ENV_VAR = "CRAFT_SHARED_CACHE"
STORE_API_ENV_VAR = "CHARMCRAFT_STORE_API_URL"
STORE_STORAGE_ENV_VAR = "CHARMCRAFT_UPLOAD_URL"
STORE_REGISTRY_ENV_VAR = "CHARMCRAFT_REGISTRY_URL"
# These are only for use within the managed environment
MANAGED_MODE_ENV_VAR = "CHARMCRAFT_MANAGED_MODE"
# endregion
# region Project files and directories
CHARMCRAFT_FILENAME = "charmcraft.yaml"
BUNDLE_FILENAME = "bundle.yaml"
MANIFEST_FILENAME = "manifest.yaml"
JUJU_CONFIG_FILENAME = "config.yaml"
METADATA_FILENAME = "metadata.yaml"
JUJU_ACTIONS_FILENAME = "actions.yaml"

WORK_DIRNAME = "work_dir"
BUILD_DIRNAME = "build"
VENV_DIRNAME = "venv"
STAGING_VENV_DIRNAME = "staging-venv"
# endregion
# region Output files and directories
# Dispatch script filename
DISPATCH_FILENAME = "dispatch"
# Hooks directory name
HOOKS_DIRNAME = "hooks"
# The minimum set of hooks to be provided for compatibility with old Juju
MANDATORY_HOOK_NAMES = frozenset(("install", "start", "upgrade-charm"))

CommonBaseStr = Literal[  # Bases supported as both build bases and run bases
    "ubuntu@18.04",
    "ubuntu@20.04",
    "ubuntu@22.04",
    "ubuntu@23.10",
    "ubuntu@24.04",
    "centos@7",
    "almalinux@9",
]
BaseStr = CommonBaseStr
BuildBaseStr = CommonBaseStr | Literal["ubuntu@devel"]

DEVEL_BASE_STRINGS = ()  # Bases that require a specified build base.

SUPPORTED_BASES = frozenset(
    (
        BaseName("ubuntu", "18.04"),
        BaseName("ubuntu", "20.04"),
        BaseName("ubuntu", "22.04"),
        BaseName("ubuntu", "23.10"),
        BaseName("ubuntu", "24.04"),
        BaseName("ubuntu", "devel"),
        BaseName("centos", "7"),
        BaseName("almalinux", "9"),
    )
)

SUPPORTED_OSES = frozenset(base.name for base in SUPPORTED_BASES)


class CharmArch(str, enum.Enum):
    """An architecture for a charm."""

    amd64 = "amd64"
    arm64 = "arm64"
    armhf = "armhf"
    ppc64el = "ppc64el"
    riscv64 = "riscv64"
    s390x = "s390x"

    def __str__(self) -> str:
        return str(self.value)


SUPPORTED_ARCHITECTURES = frozenset(arch.value for arch in CharmArch)


# The minimum set of files for a charm to be considered valid
CHARM_MANDATORY_FILES = frozenset(
    (
        DISPATCH_FILENAME,
        HOOKS_DIRNAME,
    )
)
# Optional files that can be present in a charm
CHARM_OPTIONAL_FILES = frozenset(
    (
        METADATA_FILENAME,
        JUJU_ACTIONS_FILENAME,
        JUJU_CONFIG_FILENAME,
        "metrics.yaml",
        "lxd-profile.yaml",
        "templates",
        "version",
        "lib",
        "mod",
        "LICENSE",
        "icon.svg",
        "README.md",
        "actions",
    )
)
# endregion

DEPENDENCIES_HASH_FILENAME = "charmcraft-dependencies-hash.txt"

# If Juju doesn't support the dispatch mechanism, it will execute the
# hook, and we'd need sys.argv[0] to be the name of the hook but it's
# getting lost by calling this dispatch, so we fake JUJU_DISPATCH_PATH
# to be the value it would've otherwise been.
DISPATCH_CONTENT = """#!/bin/sh

JUJU_DISPATCH_PATH="${{JUJU_DISPATCH_PATH:-$0}}" PYTHONPATH=lib:venv \\
  exec ./{entrypoint_relative_path}
"""

UBUNTU_LTS_STABLE = frozenset(
    (
        "18.04",
        "20.04",
        "22.04",
        "24.04",
    )
)

# Metadata keys that are defined in the metadata.yaml file, for backwards compatible
METADATA_YAML_KEYS = frozenset(
    (
        "assumes",
        "containers",
        "description",
        "devices",
        "display-name",
        "docs",
        "extra-bindings",
        "issues",
        "maintainers",
        "name",
        "peers",
        "provides",
        "requires",
        "resources",
        "series",
        "storage",
        "subordinate",
        "summary",
        "terms",
        "website",
    )
)

CHARM_METADATA_LEGARY_KEYS_ALIAS = frozenset(
    (
        "display_name",
        "extra_bindings",
    )
)

# Metadata keys that are allowed in the charmcraft.yaml file
CHARM_METADATA_KEYS = frozenset(
    (
        "assumes",
        "containers",
        "description",
        "devices",
        "title",
        "documentation",
        "extra-bindings",
        "links",
        "name",
        "peers",
        "provides",
        "requires",
        "resources",
        "storage",
        "subordinate",
        "summary",
        "terms",
    )
)

CHARM_METADATA_KEYS_ALIAS = frozenset(("extra_bindings",))

METADATA_YAML_MIGRATE_FIELDS = ("name", "summary", "description")
"""Fields that can exist in metadata.yaml or charmcraft.yaml, but not both."""



# Copyright 2020-2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import sys

import pydantic
import pytest

from charmcraft import parts

pytestmark = pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
# -- tests for part config processing


@pytest.mark.usefixtures("new_path")
@pytest.mark.parametrize(
    "binary_packages",
    [
        {},
        {"charm-binary-python-packages": ["pydantic-core"]},
    ],
)
@pytest.mark.parametrize("packages", [{}, {"charm-python-packages": ["pytest"]}])
@pytest.mark.parametrize("reqs", [{}, {"charm-requirements": ["requirements.lock"]}])
@pytest.mark.parametrize("strict_deps", [{}, {"charm-strict-dependencies": False}])
@pytest.mark.parametrize("entrypoint", [{}, {"charm-entrypoint": "my_charm.py"}])
def test_partconfig_happy_validation_and_completion(
    binary_packages: dict[str, str],
    packages: dict[str, str],
    reqs: dict[str, str],
    strict_deps: dict[str, bool],
    entrypoint: dict[str, str],
):
    data: dict[str, str | bool] = {
        "plugin": "charm",
        "source": ".",
    }
    data.update(binary_packages)
    data.update(packages)
    data.update(strict_deps)
    data.update(entrypoint)

    completed = parts.process_part_config(data)
    assert completed == data


def test_partconfig_no_plugin():
    data = {
        "source": ".",
    }
    with pytest.raises(ValueError) as raised:
        parts.process_part_config(data)
    assert str(raised.value) == "'plugin' not defined"


def test_partconfig_bad_property():
    data = {
        "plugin": "charm",
        "source": ".",
        "color": "purple",
    }
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.process_part_config(data)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("color",)
    assert err[0]["msg"] == "Extra inputs are not permitted"


def test_partconfig_bad_plugin_property():
    data = {
        "plugin": "charm",
        "charm-timeout": "never",
        "source": ".",
    }
    with pytest.raises(pydantic.ValidationError) as raised:
        parts.process_part_config(data)
    err = raised.value.errors()
    assert len(err) == 1
    assert err[0]["loc"] == ("charm-timeout",)
    assert err[0]["msg"] == "Extra inputs are not permitted"



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Extension application helpers."""

import copy
from pathlib import Path
from typing import Any, cast

from charmcraft.extensions.extension import Extension
from charmcraft.extensions.registry import get_extension_class


def apply_extensions(project_root: Path, yaml_data: dict[str, Any]) -> dict[str, Any]:
    """Apply all extensions.

    :param dict yaml_data: Loaded, unprocessed charmcraft.yaml
    :returns: Modified charmcraft.yaml data with extensions applied
    """
    # Don't modify the dict passed in
    yaml_data = copy.deepcopy(yaml_data)
    declared_extensions: list[str] = cast(list[str], yaml_data.get("extensions", []))
    if not declared_extensions:
        return yaml_data

    del yaml_data["extensions"]

    # Process extensions in a consistent order
    for extension_name in sorted(declared_extensions):
        extension_class = get_extension_class(extension_name)
        extension = extension_class(project_root=project_root, yaml_data=copy.deepcopy(yaml_data))
        extension.validate(extension_name=extension_name)
        _apply_extension(yaml_data, extension)
    return yaml_data


def _apply_extension(
    yaml_data: dict[str, Any],
    extension: Extension,
) -> None:
    # Apply the root components of the extension (if any)
    root_extension = extension.get_root_snippet()
    for property_name, property_value in root_extension.items():
        yaml_data[property_name] = _apply_extension_property(
            yaml_data.get(property_name), property_value
        )

    # Next, apply the part-specific components
    part_extension = extension.get_part_snippet()
    if "parts" not in yaml_data:
        yaml_data["parts"] = {}

    parts = yaml_data["parts"]
    for _, part_definition in parts.items():
        for property_name, property_value in part_extension.items():
            part_definition[property_name] = _apply_extension_property(
                part_definition.get(property_name), property_value
            )

    # Finally, add any parts specified in the extension
    parts_snippet = extension.get_parts_snippet()
    parts_names = (pn for pn in parts_snippet if pn not in yaml_data["parts"])
    for part_name in parts_names:
        parts[part_name] = parts_snippet[part_name]


def _apply_extension_property(
    existing_property: dict | list, extension_property: dict | list
) -> dict | list:
    if existing_property:
        # If the property is not scalar, merge them
        if isinstance(existing_property, list) and isinstance(extension_property, list):
            merged = extension_property + existing_property

            # If the lists are just strings, remove duplicates.
            if all(isinstance(item, str) for item in merged):
                return _remove_list_duplicates(merged)

            return merged

        if isinstance(existing_property, dict) and isinstance(extension_property, dict):
            for key, value in extension_property.items():
                existing_property[key] = _apply_extension_property(
                    existing_property.get(key), value
                )
            return existing_property
        return existing_property

    return extension_property


def _remove_list_duplicates(seq: list[str]) -> list[str]:
    """De-dupe string list maintaining ordering."""
    seen: set[str] = set()
    deduped: list[str] = []

    for item in seq:
        if item not in seen:
            seen.add(item)
            deduped.append(item)

    return deduped



# Copyright 2021-2022 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

import json
import sys
import zipfile
from argparse import ArgumentParser, Namespace

import pytest
from craft_cli import CraftError

from charmcraft import linters
from charmcraft.application.commands.analyse import Analyse
from charmcraft.cmdbase import JSON_FORMAT
from charmcraft.models.lint import LintResult


def test_options_format_possible_values(config):
    """The format option implies a set of validations."""
    cmd = Analyse(config)
    parser = ArgumentParser()
    cmd.fill_parser(parser)
    (action,) = (action for action in parser._actions if action.dest == "format")
    assert action.choices == ["json"]


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
@pytest.mark.parametrize("modebits", [0o777, 0o750, 0o444])
def test_expanded_charm_permissions(config, fake_project_dir, monkeypatch, modebits):
    """Check that the expanded charm keeps original permissions."""
    # prepare a fake charm file with some specific content just to check it was used properly
    charm_file = fake_project_dir / "foobar.charm"
    payload_file = fake_project_dir / "payload.txt"
    payload_file.write_bytes(b"123")
    payload_file.chmod(modebits)
    with zipfile.ZipFile(str(charm_file), "w") as zf:
        zf.write(str(payload_file), payload_file.name)

    args = Namespace(filepath=charm_file, force=None, format=None, ignore=None)
    Analyse(config).run(args)


@pytest.mark.skipif(sys.platform == "win32", reason="Windows not [yet] supported")
def test_corrupt_charm(fake_project_dir, config):
    """There was a problem opening the indicated charm."""
    charm_file = fake_project_dir / "foobar.charm"
    charm_file.write_text("this is not a real zip content")

    args = Namespace(filepath=charm_file, force=None, format=None, ignore=None)
    with pytest.raises(CraftError) as cm:
        Analyse(config).run(args)
    assert str(cm.value) == (f"Cannot open charm file '{charm_file}': File is not a zip file")


def create_a_valid_zip(tmp_path):
    """Prepare a simple zip file."""
    zip_file = tmp_path / "foobar.charm"
    with zipfile.ZipFile(str(zip_file), "w") as zf:
        zf.writestr("fake_file", b"fake content")
    return zip_file


def test_integration_linters(fake_project_dir, emitter, config, monkeypatch):
    """Integration test with a real analysis."""
    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    Analyse(config).run(args)

    emitter.assert_progress(
        "language: Charm language unknown (https://juju.is/docs/sdk/charmcraft-analyzers-and-linters#heading--language)",
        permanent=True,
    )


@pytest.mark.parametrize("indicated_format", [None, JSON_FORMAT])
def test_complete_set_of_results(
    check, emitter, service_factory, config, monkeypatch, fake_project_dir, indicated_format
):
    """Show a complete basic case of results."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-lint-01",
            check_type=linters.CheckType.LINT,
            url="url-01",
            text="text-01",
            result=LintResult.WARNING,
        ),
        linters.CheckResult(
            name="check-lint-02",
            check_type=linters.CheckType.LINT,
            url="url-02",
            text="text-02",
            result=LintResult.OK,
        ),
        linters.CheckResult(
            name="check-lint-03",
            check_type=linters.CheckType.LINT,
            url="url-03",
            text="text-03",
            result=LintResult.ERROR,
        ),
        linters.CheckResult(
            name="check-attribute-04",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url-04",
            text="text-04",
            result="check-result-04",
        ),
        linters.CheckResult(
            name="check-attribute-05",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url-05",
            text="text-05",
            result=LintResult.IGNORED,
        ),
        linters.CheckResult(
            name="check-lint-06",
            check_type=linters.CheckType.LINT,
            url="url-06",
            text="text-06",
            result=LintResult.IGNORED,
        ),
        linters.CheckResult(
            name="check-lint-07",
            check_type=linters.CheckType.LINT,
            url="url-07",
            text="text-07",
            result=LintResult.FATAL,
        ),
    ]

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=indicated_format, ignore=None)
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )
    Analyse(config).run(args)

    if indicated_format is None:
        expected = [
            "check-lint-01: [WARNING] text-01 (url-01)",
            "check-lint-02: [OK] text-02 (url-02)",
            "check-lint-03: [ERROR] text-03 (url-03)",
            "check-attribute-04: [CHECK-RESULT-04] text-04 (url-04)",
            "check-attribute-05: (url-05) ",
            "check-lint-06: (url-06) ",
            "check-lint-07: [FATAL] text-07 (url-07)",
        ]
        for line in expected:
            with check:
                emitter.assert_progress(line, permanent=True)
    else:
        expected = [
            {
                "check_type": "lint",
                "name": "check-lint-01",
                "result": "warning",
                "text": "text-01",
                "url": "url-01",
            },
            {
                "check_type": "lint",
                "name": "check-lint-02",
                "result": "ok",
                "text": "text-02",
                "url": "url-02",
            },
            {
                "check_type": "lint",
                "name": "check-lint-03",
                "result": "error",
                "text": "text-03",
                "url": "url-03",
            },
            {
                "check_type": "attribute",
                "name": "check-attribute-04",
                "result": "check-result-04",
                "text": "text-04",
                "url": "url-04",
            },
            {
                "check_type": "attribute",
                "name": "check-attribute-05",
                "result": "ignored",
                "text": "text-05",
                "url": "url-05",
            },
            {
                "check_type": "lint",
                "name": "check-lint-06",
                "result": "ignored",
                "text": "text-06",
                "url": "url-06",
            },
            {
                "check_type": "lint",
                "name": "check-lint-07",
                "result": "fatal",
                "text": "text-07",
                "url": "url-07",
            },
        ]
        text = emitter.assert_message(r"\[.*\]", regex=True)
        assert expected == json.loads(text)


def test_only_attributes(emitter, service_factory, config, monkeypatch, fake_project_dir):
    """Show only attribute results (the rest may be ignored)."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-attribute",
            check_type=linters.CheckType.ATTRIBUTE,
            url="url",
            text="text",
            result="check-result",
        ),
    ]

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )
    retcode = Analyse(config).run(args)

    emitter.assert_progress("check-attribute: [CHECK-RESULT] text (url)", permanent=True)
    assert retcode == 0


def test_only_warnings(emitter, service_factory, config, monkeypatch, fake_project_dir):
    """Show only warning results (the rest may be ignored)."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-lint",
            check_type=linters.CheckType.LINT,
            url="url",
            text="text",
            result=LintResult.WARNING,
        ),
    ]

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )
    retcode = Analyse(config).run(args)

    emitter.assert_progress("check-lint: [WARNING] text (url)", permanent=True)
    assert retcode == 3


def test_only_errors(emitter, service_factory, config, monkeypatch, fake_project_dir):
    """Show only error results (the rest may be ignored)."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-lint",
            check_type=linters.CheckType.LINT,
            url="url",
            text="text",
            result=LintResult.ERROR,
        ),
    ]

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )
    retcode = Analyse(config).run(args)

    emitter.assert_progress("check-lint: [ERROR] text (url)", permanent=True)
    assert retcode == 2


def test_both_errors_and_warnings(emitter, service_factory, config, monkeypatch, fake_project_dir):
    """Show error and warnings results."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-lint-1",
            check_type=linters.CheckType.LINT,
            url="url-1",
            text="text-1",
            result=LintResult.ERROR,
        ),
        linters.CheckResult(
            name="check-lint-2",
            check_type=linters.CheckType.LINT,
            url="url-2",
            text="text-2",
            result=LintResult.WARNING,
        ),
    ]

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )
    retcode = Analyse(config).run(args)

    emitter.assert_progress("check-lint-1: [ERROR] text-1 (url-1)", permanent=True)
    emitter.assert_progress("check-lint-2: [WARNING] text-2 (url-2)", permanent=True)
    assert retcode == 2


def test_only_lint_ok(emitter, service_factory, config, monkeypatch, fake_project_dir):
    """Show only lint results that are ok (the rest may be ignored)."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-lint",
            check_type=linters.CheckType.LINT,
            url="url",
            text="text",
            result=LintResult.OK,
        ),
    ]

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )
    retcode = Analyse(config).run(args)

    emitter.assert_progress("check-lint: [OK] text (url)", permanent=True)
    assert retcode == 0


def test_only_fatal(emitter, service_factory, config, monkeypatch, fake_project_dir):
    """Show only fatal lint results (the rest may be ignored)."""
    # fake results from the analyzer
    linting_results = [
        linters.CheckResult(
            name="check-lint",
            check_type=linters.CheckType.LINT,
            url="url",
            text="text",
            result=LintResult.FATAL,
        ),
    ]
    monkeypatch.setattr(
        service_factory.analysis, "lint_directory", lambda *a, **k: linting_results
    )

    fake_charm = create_a_valid_zip(fake_project_dir)
    args = Namespace(filepath=fake_charm, force=None, format=None, ignore=None)
    retcode = Analyse(config).run(args)

    emitter.assert_progress("check-lint: [FATAL] text (url)", permanent=True)
    assert retcode == 1



# Copyright 2023-2024 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""General purpose library functions for charmcraft."""

import ast
import hashlib
import os
import pathlib
from dataclasses import dataclass
from typing import overload

import yaml
from craft_cli import CraftError

from charmcraft import const, errors


@dataclass(frozen=True)
class LibData:
    """All data fields for a library, including external ones."""

    lib_id: str | None
    api: int
    patch: int
    content: str | None
    content_hash: str | None
    full_name: str
    path: pathlib.Path
    lib_name: str
    charm_name: str


@dataclass
class LibInternals:
    """All internals from a lib: the metadata fields, the hash and the content itself."""

    lib_id: str
    api: int
    patch: int
    pydeps: list[str]
    content_hash: str
    content: str


def get_name_from_metadata() -> str | None:
    """Return the name if present and plausible in metadata.yaml."""
    try:
        with open(const.METADATA_FILENAME, "rb") as fh:
            metadata = yaml.safe_load(fh)
        charm_name = metadata["name"]
    except (yaml.error.YAMLError, OSError, KeyError):
        return None
    return charm_name


def create_importable_name(charm_name: str) -> str:
    """Convert a charm name to something that is importable in python."""
    return charm_name.replace("-", "_")


def create_charm_name_from_importable(charm_name: str) -> str:
    """Convert a charm name from the importable form to the real form."""
    # _ is invalid in charm names, so we know it's intended to be '-'
    return charm_name.replace("_", "-")


def get_lib_internals(lib_path: pathlib.Path) -> LibInternals:
    """Get content, its hash, and all the metadata fields from a library.

    There are two kind of metadata fields, simple constant fields and PYDEPS. The simple
    constant fields are also the mandatory ones: LIBAPI, LIBPATCH and LIBID.
    """
    content = lib_path.read_text(encoding="utf8")
    try:
        tree = ast.parse(content)
    except Exception:
        raise CraftError(f"Failed to parse Python library {str(lib_path)!r}")

    def _api_patch_validator(value):
        return isinstance(value, int) and value >= 0

    _msg_prefix = f"Library {str(lib_path)!r} metadata field "
    simple_fields = {
        "LIBAPI": (
            _api_patch_validator,
            _msg_prefix + "LIBAPI must be a constant assignment of zero or a positive integer.",
        ),
        "LIBPATCH": (
            _api_patch_validator,
            _msg_prefix + "LIBPATCH must be a constant assignment of zero or a positive integer.",
        ),
        "LIBID": (
            lambda value: isinstance(value, str) and value and value.isascii(),
            _msg_prefix + "LIBID must be a constant assignment of a non-empty ASCII string.",
        ),
    }
    pydeps_error = _msg_prefix + "PYDEPS must be a constant list of non-empty strings"

    # walk the AST "first layer", find assignments to the key fields and validate them
    metadata = {}
    for node in ast.iter_child_nodes(tree):
        if isinstance(node, ast.Assign):
            for target in (n for n in node.targets if isinstance(n, ast.Name)):
                if target.id in simple_fields:
                    validator, error_msg = simple_fields[target.id]
                    if not isinstance(node.value, ast.Constant):
                        raise CraftError(error_msg)
                    real_value = node.value.value
                    if not validator(real_value):
                        raise CraftError(error_msg)
                    metadata[target.id] = real_value
                elif target.id == "PYDEPS":
                    if not isinstance(node.value, ast.List):
                        raise CraftError(pydeps_error)
                    metadata["PYDEPS"] = pydeps = []
                    for item in node.value.elts:
                        if not isinstance(item, ast.Constant):
                            raise CraftError(pydeps_error)
                        value = item.value
                        if not isinstance(value, str) or not value:
                            raise CraftError(pydeps_error)
                        pydeps.append(value)

    # extra verifications for cases that need to consider more than the individual fields
    mandatory_missing = set(simple_fields) - set(metadata)
    if mandatory_missing:
        joined = ", ".join(sorted(mandatory_missing))
        raise CraftError(
            f"Library {str(lib_path)!r} is missing the mandatory metadata fields: {joined}."
        )
    if metadata["LIBAPI"] == 0 and metadata["LIBPATCH"] == 0:
        raise CraftError(
            f"Library {str(lib_path)!r} metadata fields LIBAPI and LIBPATCH cannot both be zero."
        )

    # hash the file excluding those lines where appear keys used for version control
    metadata_vcs_fields = (b"LIBAPI", b"LIBPATCH", b"LIBID")
    hasher = hashlib.sha256()
    with lib_path.open("rb") as fh:
        for line in fh:
            # always use \n as newline for users to have the same
            # file hash both in Linux and Windows
            line = line.replace(b"\r\n", b"\n")
            if not line.startswith(metadata_vcs_fields):
                hasher.update(line)
    content_hash = hasher.hexdigest()

    return LibInternals(
        lib_id=metadata["LIBID"],
        api=metadata["LIBAPI"],
        patch=metadata["LIBPATCH"],
        pydeps=metadata.get("PYDEPS", []),
        content_hash=content_hash,
        content=content,
    )


def get_lib_path(charm: str, lib_name: str, api: int) -> pathlib.Path:
    """Get a relative path for a library based on its home charm, name and API version.

    :param charm: The name of the charm that owns this library
    :param lib_name: The name of the library
    :param api: The API version of the library
    :returns: A relative path to the library python file.
    """
    return (
        pathlib.Path("lib/charms") / create_importable_name(charm) / f"v{api}" / f"{lib_name}.py"
    )


def get_lib_module_name(charm: str, lib_name: str, api: int) -> str:
    """Get a Python module path for a library based on its home charm, name and API version.

    :param charm: The name of the charm that owns this library
    :param lib_name: The name of the library
    :param api: The API version of the library
    :returns: A string of the full path to the charm.
    """
    return f"charms.{create_importable_name(charm)}.v{api}.{lib_name}"


@overload
def get_lib_info(*, full_name: str) -> LibData: ...
@overload
def get_lib_info(*, lib_path: pathlib.Path) -> LibData: ...
def get_lib_info(*, full_name: str | None = None, lib_path: pathlib.Path | None = None) -> LibData:
    """Get the whole lib info from the path/file.

    This will perform mutation of the charm name to create importable paths.
    * `charm_name` and `libdata.charm_name`: `foo-bar`
    * `full_name` and `libdata.full_name`: `charms.foo_bar.v0.somelib`
    * paths, including `libdata.path`: `lib/charms/foo_bar/v0/somelib`

    This function needs to be called standing on the root directory of the project.
    """
    if lib_path:
        # get it from the lib_path
        try:
            libsdir, charmsdir, importable_charm_name, v_api = lib_path.parts[:-1]
        except ValueError:
            raise errors.BadLibraryPathError(lib_path)
        if libsdir != "lib" or charmsdir != "charms" or lib_path.suffix != ".py":
            raise errors.BadLibraryPathError(lib_path)
        full_name = ".".join((charmsdir, importable_charm_name, v_api, lib_path.stem))
    elif full_name:
        # build the path! convert a lib name with dots to the full path, including lib
        # dir and Python extension.
        #    e.g.: charms.mycharm.v4.foo -> lib/charms/mycharm/v4/foo.py
        try:
            charmsdir, charm_name, v_api, libfile = full_name.split(".")
        except ValueError:
            raise errors.BadLibraryNameError(full_name)

        # the lib full_name includes the charm_name which might not be importable (dashes)
        importable_charm_name = create_importable_name(charm_name)

        if charmsdir != "charms":
            raise errors.BadLibraryNameError(full_name)
        path = pathlib.Path("lib")
        lib_path = path / charmsdir / importable_charm_name / v_api / (libfile + ".py")
    else:
        raise TypeError("get_lib_info needs either a full name or a lib path")

    # charm names in the path can contain '_' to be importable
    # these should be '-', so change them back
    charm_name = create_charm_name_from_importable(importable_charm_name)

    if v_api[0] != "v" or not v_api[1:].isdigit():
        raise CraftError("The API version in the library path must be 'vN' where N is an integer.")
    api_from_path = int(v_api[1:])

    lib_name = lib_path.stem
    if not lib_path.exists():
        return LibData(
            lib_id=None,
            api=api_from_path,
            patch=-1,
            content_hash=None,
            content=None,
            full_name=full_name,
            path=lib_path,
            lib_name=lib_name,
            charm_name=charm_name,
        )

    internals = get_lib_internals(lib_path)

    # validate internal API matches with what was used in the path
    if internals.api != api_from_path:
        raise CraftError(
            f"Library {str(lib_path)!r} metadata field LIBAPI is different from the version in the path."
        )

    return LibData(
        lib_id=internals.lib_id,
        api=internals.api,
        patch=internals.patch,
        content_hash=internals.content_hash,
        content=internals.content,
        full_name=full_name,
        path=lib_path,
        lib_name=lib_name,
        charm_name=charm_name,
    )


def get_libs_from_tree(
    charm_name: str | None = None, root: pathlib.Path | None = None
) -> list[LibData]:
    """Get library info from the directories tree (for a specific charm if specified).

    It only follows/uses the directories/files for a correct charmlibs
    disk structure.

    This can take charm_name as both importable and normal form.
    """
    local_libs_data = []

    current_directory = os.getcwd()
    if root is not None:
        os.chdir(root)
    try:
        if charm_name is None:
            base_dir = pathlib.Path("lib") / "charms"
            charm_dirs = sorted(base_dir.iterdir()) if base_dir.is_dir() else []
        else:
            importable_charm_name = create_importable_name(charm_name)
            base_dir = pathlib.Path("lib") / "charms" / importable_charm_name
            charm_dirs = [base_dir] if base_dir.is_dir() else []

        for charm_dir in charm_dirs:
            for v_dir in sorted(charm_dir.iterdir()):
                if v_dir.is_dir() and v_dir.name[0] == "v" and v_dir.name[1:].isdigit():
                    for libfile in sorted(v_dir.glob("*.py")):
                        local_libs_data.append(get_lib_info(lib_path=libfile))
    finally:
        os.chdir(current_directory)
    return local_libs_data


def collect_charmlib_pydeps(basedir: pathlib.Path) -> set[str]:
    """Collect the Python dependencies from all the project's charm libraries."""
    all_libs_data = get_libs_from_tree(root=basedir)
    charmlib_deps = set()
    for libdata in all_libs_data:
        internals = get_lib_internals(basedir / libdata.path)
        charmlib_deps.update(internals.pydeps)
    return charmlib_deps



#!/usr/bin/env python3

import argparse
import os
import pathlib
import sys

from craft_cli.dispatcher import Dispatcher, _CustomArgumentParser

this_dir = pathlib.Path(os.path.split(__file__)[0])
sys.path.insert(0, str((this_dir / "..").absolute()))

from charmcraft import application, services


def command_page_header(cmd, options_str, required_str):
    underline = "=" * len(cmd.name)
    overview = fix_spelling(cmd.overview)
    return f"""
.. _ref_commands_{cmd.name}:

{cmd.name}
{underline}
{overview}

Usage
-----

:command:`charmcraft {cmd.name}{options_str}{required_str}`

"""


# Temporary fix for spelling checks.
def fix_spelling(t):
    for old, new in [("artifact", "artefact"), ("Initialize", "Initialise")]:
        t = t.replace(old, new)
    return t


def make_sentence(t):
    return t[:1].upper() + t[1:].rstrip(".") + "."


def not_none(*args):
    return [x for x in args if x != None]


def make_section(title, items):
    s = ""
    if items:
        underline = "-" * len(title)
        s = f"{title}\n{underline}\n\n"

    for dest, (names, help_str) in sorted(items):
        names = " or ".join([f"``{name}``" for name in names])
        description = make_sentence(fix_spelling(help_str))
        s += f"{names}\n   {description}\n"

    s += "\n"
    return s


def main(docs_dir):
    """Generate reference documentation for the command line interface,
    creating pages in the docs/reference/commands directory and creating the
    directory itself if necessary."""

    # Create the directory for the commands reference.
    commands_ref_dir = docs_dir / "reference" / "commands"
    if not commands_ref_dir.exists():
        commands_ref_dir.mkdir()

    # Create a dispatcher like Charmcraft does to get access to the same options.
    charmcraft_services = services.CharmcraftServiceFactory(app=application.APP_METADATA)
    app = application.Charmcraft(app=application.APP_METADATA, services=charmcraft_services)
    application.commands.fill_command_groups(app)
    command_groups = app.command_groups

    dispatcher = Dispatcher(
        app.app.name,
        command_groups,
        summary=str(app.app.summary),
        extra_global_args=app._global_arguments,
    )

    help_builder = dispatcher._help_builder

    global_options = {}
    for arg in dispatcher.global_arguments:
        opts = not_none(arg.short_option, arg.long_option)
        global_options[arg.name] = (opts, arg.help_message)

    toc = []

    for group in command_groups:
        group_name = group.name.lower() + "-commands" + os.extsep + "rst"
        group_path = commands_ref_dir / group_name
        g = group_path.open("w")

        for cmd_class in group.commands:
            if cmd_class.hidden:
                continue
            cmd = cmd_class({"app": {}, "services": {}})
            p = _CustomArgumentParser(help_builder)
            cmd.fill_parser(p)

            options = {}
            required = []
            options_str = " "

            for action in sorted(p._actions, key=lambda a: a.dest):
                if action.required:
                    required.append((action.dest, ([action.metavar], action.help)))
                elif action.option_strings and action.dest not in global_options:
                    options[action.dest] = (action.option_strings, action.help)

            cmd_path = commands_ref_dir / (cmd.name + os.extsep + "rst")

            if options or global_options:
                options_str += "[options]"
            required_str = "".join([(" <%s>" % vl[0]) for d, (vl, h) in required])

            f = cmd_path.open("w")
            f.write(command_page_header(cmd, options_str, required_str))
            f.write(make_section("Required", required))
            f.write(make_section("Options", options.items()))
            f.write(make_section("Global options", global_options.items()))

            # Add a section for the command to be included in the group reference.
            g.write(f":ref:`ref_commands_{cmd.name}`\n")
            g.write(
                "   " + make_sentence(fix_spelling(cmd.help_msg)).replace("\n", "\n   ") + "\n\n"
            )

            # Add an entry in the table of contents.
            toc.append(cmd.name)

    toc_path = commands_ref_dir / "toc.rst"
    f = toc_path.open("w")
    f.write(".. toctree::\n   :hidden:\n\n")
    for name in sorted(toc):
        f.write(f"   /reference/commands/{name}\n")


if __name__ == "__main__":
    if len(sys.argv) == 2:
        main(pathlib.Path(sys.argv[1]))



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
import pytest

from charmcraft.errors import ExtensionError
from charmcraft.extensions import apply_extensions
from charmcraft.extensions.app import (
    DjangoFramework,
    FlaskFramework,
    GoFramework,
)


def make_flask_input_yaml():
    return {
        "type": "charm",
        "name": "test-flask",
        "summary": "test summary",
        "description": "test description",
        "bases": [{"name": "ubuntu", "channel": "22.04"}],
        "extensions": ["flask-framework"],
    }


@pytest.fixture(name="flask_input_yaml")
def flask_input_yaml_fixture():
    return make_flask_input_yaml()


@pytest.mark.parametrize(
    ("input_yaml", "experimental", "expected"),
    [
        (
            make_flask_input_yaml(),
            False,
            {
                "actions": FlaskFramework.actions,
                "assumes": ["k8s-api"],
                "bases": [{"channel": "22.04", "name": "ubuntu"}],
                "containers": {
                    "flask-app": {"resource": "flask-app-image"},
                },
                "description": "test description",
                "name": "test-flask",
                "charm-libs": [
                    {"lib": "traefik_k8s.ingress", "version": "2"},
                    {"lib": "observability_libs.juju_topology", "version": "0"},
                    {"lib": "grafana_k8s.grafana_dashboard", "version": "0"},
                    {"lib": "loki_k8s.loki_push_api", "version": "0"},
                    {"lib": "data_platform_libs.data_interfaces", "version": "0"},
                    {"lib": "prometheus_k8s.prometheus_scrape", "version": "0"},
                    {"lib": "redis_k8s.redis", "version": "0"},
                    {"lib": "data_platform_libs.s3", "version": "0"},
                    {"lib": "saml_integrator.saml", "version": "0"},
                ],
                "config": {
                    "options": {**FlaskFramework.options},
                },
                "parts": {
                    "charm": {
                        "plugin": "charm",
                        "source": ".",
                        "build-snaps": ["rustup"],
                        "override-build": "rustup default stable\ncraftctl default",
                    }
                },
                "peers": {"secret-storage": {"interface": "secret-storage"}},
                "provides": {
                    "metrics-endpoint": {"interface": "prometheus_scrape"},
                    "grafana-dashboard": {"interface": "grafana_dashboard"},
                },
                "requires": {
                    "logging": {"interface": "loki_push_api"},
                    "ingress": {"interface": "ingress", "limit": 1},
                },
                "resources": {
                    "flask-app-image": {
                        "description": "flask application image.",
                        "type": "oci-image",
                    },
                },
                "summary": "test summary",
                "type": "charm",
            },
        ),
        (
            {
                "type": "charm",
                "name": "test-django",
                "summary": "test summary",
                "description": "test description",
                "base": "ubuntu@22.04",
                "platforms": {
                    "amd64": None,
                    "arm64": None,
                    "armhf": None,
                    "ppc64el": None,
                    "riscv64": None,
                    "s390x": None,
                },
                "extensions": ["django-framework"],
            },
            True,
            {
                "actions": DjangoFramework.actions,
                "assumes": ["k8s-api"],
                "base": "ubuntu@22.04",
                "platforms": {
                    "amd64": None,
                    "arm64": None,
                    "armhf": None,
                    "ppc64el": None,
                    "riscv64": None,
                    "s390x": None,
                },
                "containers": {
                    "django-app": {"resource": "django-app-image"},
                },
                "description": "test description",
                "name": "test-django",
                "charm-libs": [
                    {"lib": "traefik_k8s.ingress", "version": "2"},
                    {"lib": "observability_libs.juju_topology", "version": "0"},
                    {"lib": "grafana_k8s.grafana_dashboard", "version": "0"},
                    {"lib": "loki_k8s.loki_push_api", "version": "0"},
                    {"lib": "data_platform_libs.data_interfaces", "version": "0"},
                    {"lib": "prometheus_k8s.prometheus_scrape", "version": "0"},
                    {"lib": "redis_k8s.redis", "version": "0"},
                    {"lib": "data_platform_libs.s3", "version": "0"},
                    {"lib": "saml_integrator.saml", "version": "0"},
                ],
                "config": {
                    "options": {**DjangoFramework.options},
                },
                "parts": {
                    "charm": {
                        "plugin": "charm",
                        "source": ".",
                        "build-snaps": ["rustup"],
                        "override-build": "rustup default stable\ncraftctl default",
                    }
                },
                "peers": {"secret-storage": {"interface": "secret-storage"}},
                "provides": {
                    "metrics-endpoint": {"interface": "prometheus_scrape"},
                    "grafana-dashboard": {"interface": "grafana_dashboard"},
                },
                "requires": {
                    "logging": {"interface": "loki_push_api"},
                    "ingress": {"interface": "ingress", "limit": 1},
                },
                "resources": {
                    "django-app-image": {
                        "description": "django application image.",
                        "type": "oci-image",
                    },
                },
                "summary": "test summary",
                "type": "charm",
            },
        ),
        (
            {
                "type": "charm",
                "name": "test-go",
                "summary": "test summary",
                "description": "test description",
                "bases": [{"name": "ubuntu", "channel": "24.04"}],
                "extensions": ["go-framework"],
            },
            True,
            {
                "actions": GoFramework.actions,
                "assumes": ["k8s-api"],
                "bases": [{"channel": "24.04", "name": "ubuntu"}],
                "containers": {
                    "app": {"resource": "app-image"},
                },
                "description": "test description",
                "name": "test-go",
                "charm-libs": [
                    {"lib": "traefik_k8s.ingress", "version": "2"},
                    {"lib": "observability_libs.juju_topology", "version": "0"},
                    {"lib": "grafana_k8s.grafana_dashboard", "version": "0"},
                    {"lib": "loki_k8s.loki_push_api", "version": "0"},
                    {"lib": "data_platform_libs.data_interfaces", "version": "0"},
                    {"lib": "prometheus_k8s.prometheus_scrape", "version": "0"},
                    {"lib": "redis_k8s.redis", "version": "0"},
                    {"lib": "data_platform_libs.s3", "version": "0"},
                    {"lib": "saml_integrator.saml", "version": "0"},
                ],
                "config": {
                    "options": {**GoFramework.options},
                },
                "parts": {
                    "charm": {
                        "plugin": "charm",
                        "source": ".",
                        "build-snaps": ["rustup"],
                        "override-build": "rustup default stable\ncraftctl default",
                    }
                },
                "peers": {"secret-storage": {"interface": "secret-storage"}},
                "provides": {
                    "metrics-endpoint": {"interface": "prometheus_scrape"},
                    "grafana-dashboard": {"interface": "grafana_dashboard"},
                },
                "requires": {
                    "logging": {"interface": "loki_push_api"},
                    "ingress": {"interface": "ingress", "limit": 1},
                },
                "resources": {
                    "app-image": {
                        "description": "go application image.",
                        "type": "oci-image",
                    },
                },
                "summary": "test summary",
                "type": "charm",
            },
        ),
    ],
)
def test_apply_extensions_correct(monkeypatch, experimental, tmp_path, input_yaml, expected):
    if experimental:
        monkeypatch.setenv("CHARMCRAFT_ENABLE_EXPERIMENTAL_EXTENSIONS", "1")

    applied = apply_extensions(tmp_path, input_yaml)
    assert applied == expected


PROTECTED_FIELDS_TEST_PARAMETERS = [
    pytest.param({"type": "bundle"}, id="type"),
    pytest.param({"containers": {"foobar": {"resource": "foobar"}}}, id="containers"),
    pytest.param({"peers": {"foobar": {"interface": "foobar"}}}, id="peers"),
    pytest.param({"resources": {"foobar": {"type": "oci-image"}}}, id="resources"),
]


@pytest.mark.parametrize("modification", PROTECTED_FIELDS_TEST_PARAMETERS)
def test_flask_protected_fields(modification, flask_input_yaml, tmp_path):
    flask_input_yaml.update(modification)
    with pytest.raises(ExtensionError):
        apply_extensions(tmp_path, flask_input_yaml)


def test_flask_merge_options(flask_input_yaml, tmp_path):
    added_options = {"api_secret": {"type": "string"}}
    flask_input_yaml["config"] = {"options": added_options}
    applied = apply_extensions(tmp_path, flask_input_yaml)
    assert applied["config"] == {
        "options": {
            **FlaskFramework.options,
            **added_options,
        }
    }


def test_flask_merge_action(flask_input_yaml, tmp_path):
    added_actions = {"foobar": {}}
    flask_input_yaml["actions"] = added_actions
    applied = apply_extensions(tmp_path, flask_input_yaml)
    assert applied["actions"] == {**FlaskFramework.actions, **added_actions}


def test_flask_merge_relation(flask_input_yaml, tmp_path):
    new_provides = {"provides-foobar": {"interface": "foobar"}}
    new_requires = {"requires-foobar": {"interface": "foobar"}}
    flask_input_yaml["provides"] = new_provides
    flask_input_yaml["requires"] = new_requires
    applied = apply_extensions(tmp_path, flask_input_yaml)
    assert applied["provides"] == {
        "metrics-endpoint": {"interface": "prometheus_scrape"},
        "grafana-dashboard": {"interface": "grafana_dashboard"},
        **new_provides,
    }
    assert applied["requires"] == {
        "logging": {"interface": "loki_push_api"},
        "ingress": {"interface": "ingress", "limit": 1},
        **new_requires,
    }


def test_flask_merge_charm_libs(flask_input_yaml, tmp_path):
    added_charm_libs = [{"lib": "smtp_integrator.smtp", "version": "0"}]
    flask_input_yaml["charm-libs"] = added_charm_libs
    applied = apply_extensions(tmp_path, flask_input_yaml)
    assert applied["charm-libs"] == [*FlaskFramework._CHARM_LIBS, *added_charm_libs]


INCOMPATIBLE_FIELDS_TEST_PARAMETERS = [
    pytest.param({"devices": {"gpu": {"type": "gpu"}}}, id="devices"),
    pytest.param({"extra-bindings": {"foobar": {}}}, id="extra-bindings"),
    pytest.param({"storage": {"foobar": {"type": "filesystem"}}}, id="storage"),
    pytest.param(
        {"config": {"options": {"webserver-wsgi-path": {"type": "string"}}}},
        id="duplicate-options",
    ),
    pytest.param(
        {"requires": {"ingress": {"interface": "ingress"}}},
        id="duplicate-requires",
    ),
    pytest.param(
        {"provides": {"metrics-endpoint": {"interface": "prometheus_scrape"}}},
        id="duplicate-provides",
    ),
    pytest.param(
        {"config": {"options": {"webserver-foobar": {"type": "string"}}}},
        id="reserved-config-prefix-webserver",
    ),
    pytest.param(
        {"config": {"options": {"flask-foobar": {"type": "string"}}}},
        id="reserved-config-prefix-flask",
    ),
]


@pytest.mark.parametrize("modification", INCOMPATIBLE_FIELDS_TEST_PARAMETERS)
def test_flask_incompatible_fields(modification, flask_input_yaml, tmp_path):
    flask_input_yaml.update(modification)
    with pytest.raises(ExtensionError):
        apply_extensions(tmp_path, flask_input_yaml)


def test_handle_charm_part(flask_input_yaml, tmp_path):
    # Currently, in the flask-framework extension, we will reject any project that
    # includes a charm part. This is to prevent issues where a non-default charm part is
    # incompatible with this extension. This might change in the future.
    # For the same reason, the Flask-Framework extension will also add a default charm part.
    flask_input_yaml["parts"] = {"charm": {}}
    with pytest.raises(ExtensionError):
        apply_extensions(tmp_path, flask_input_yaml)
    del flask_input_yaml["parts"]
    applied = apply_extensions(tmp_path, flask_input_yaml)
    assert applied["parts"]["charm"] == {
        "plugin": "charm",
        "source": ".",
        "build-snaps": ["rustup"],
        "override-build": "rustup default stable\ncraftctl default",
    }



# Copyright 2021 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft


from unittest.mock import patch

import pytest

from charmcraft.bases import check_if_base_matches_host, get_host_as_base
from charmcraft.models.charmcraft import Base
from charmcraft.utils import OSPlatform


@pytest.fixture
def mock_get_os_platform():
    os_platform = OSPlatform(system="host-OS", release="host-CHANNEL", machine="host-ARCH")
    with patch("charmcraft.bases.get_os_platform", return_value=os_platform) as mock_platform:
        yield mock_platform


@pytest.fixture
def mock_get_host_architecture():
    with patch(
        "craft_application.util.get_host_architecture", return_value="host-ARCH"
    ) as mock_host_arch:
        yield mock_host_arch


def test_get_host_as_base(mock_get_os_platform, mock_get_host_architecture):
    assert get_host_as_base() == Base(
        name="host-os",
        channel="host-CHANNEL",
        architectures=["host-ARCH"],
    )


def test_check_if_bases_matches_host_matches(mock_get_os_platform, mock_get_host_architecture):
    base = Base(name="host-os", channel="host-CHANNEL", architectures=["host-ARCH"])
    assert check_if_base_matches_host(base) == (True, None)

    base = Base(
        name="host-os",
        channel="host-CHANNEL",
        architectures=["other-ARCH", "host-ARCH"],
    )
    assert check_if_base_matches_host(base) == (True, None)


def test_check_if_bases_matches_host_name_mismatch(
    mock_get_os_platform, mock_get_host_architecture
):
    base = Base(name="test-other-os", channel="host-CHANNEL", architectures=["host-ARCH"])

    assert check_if_base_matches_host(base) == (
        False,
        "name 'test-other-os' does not match host 'host-os'",
    )


def test_check_if_bases_matches_host_channel_mismatch(
    mock_get_os_platform, mock_get_host_architecture
):
    base = Base(name="host-os", channel="other-CHANNEL", architectures=["host-ARCH"])

    assert check_if_base_matches_host(base) == (
        False,
        "channel 'other-CHANNEL' does not match host 'host-CHANNEL'",
    )


def test_check_if_bases_matches_host_arch_mismatch(
    mock_get_os_platform, mock_get_host_architecture
):
    base = Base(
        name="host-os",
        channel="host-CHANNEL",
        architectures=["other-ARCH", "other-ARCH2"],
    )

    assert check_if_base_matches_host(base) == (
        False,
        "host architecture 'host-ARCH' not in base architectures ['other-ARCH', 'other-ARCH2']",
    )



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Infrastructure for the 'extensions' command."""
import argparse
from textwrap import dedent

from craft_cli import emit

from charmcraft import extensions, utils
from charmcraft.application.commands import base


class ListExtensionsCommand(base.CharmcraftCommand):
    """List available extensions for all supported bases."""

    common = True
    name = "list-extensions"
    help_msg = "List available extensions for all supported bases"
    overview = dedent(
        """
        List available extensions and their corresponding bases.
        """
    )
    format_option = True

    def run(self, parsed_args: argparse.Namespace):
        """Print the list of available extensions and their bases."""
        extension_data = extensions.registry.get_extensions()

        if not parsed_args.format:
            extension_data = [
                {
                    "Extension name": ext["name"],
                    "Supported bases": ", ".join(ext["bases"]),
                    "Experimental bases": ", ".join(ext["experimental_bases"]),
                }
                for ext in extension_data
            ]

        emit.message(
            utils.format_content(
                extension_data,
                fmt=parsed_args.format or utils.OutputFormat.TABLE,
            )
        )


class ExtensionsCommand(ListExtensionsCommand):
    """A command alias to list the available extensions."""

    common = False
    name = "extensions"
    hidden = True


class ExpandExtensionsCommand(base.CharmcraftCommand):
    """Expand the extensions in the charmcraft.yaml file."""

    common = True
    name = "expand-extensions"
    help_msg = "Expand extensions in charmcraft.yaml"
    overview = dedent(
        """
        Expand charmcraft.yaml using the extensions specified in the file and
        output the resulting configuration to the terminal.

        This allows you to see how the extensions used modify your existing
        charmcraft.yaml file.
        """
    )
    always_load_project = True

    def fill_parser(self, parser) -> None:
        """Fill in the parser for this command."""
        super().fill_parser(parser)

    def run(self, parsed_args: argparse.Namespace) -> None:
        """Print the project's specification with the extensions expanded."""
        emit.message(utils.dump_yaml(self._services.project.marshal()))



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft

"""Charmcraft basic pydantic model."""
from typing import Annotated

import craft_parts.constraints
import pydantic


def _validate_attribute_name(value: str) -> str:
    """Validate attribute name."""
    from charmcraft import linters  # import here to avoid cyclic imports

    valid_names = [
        checker.name
        for checker in linters.CHECKERS
        if checker.check_type == linters.CheckType.ATTRIBUTE
    ]
    if value not in valid_names:
        raise ValueError(f"Bad attribute name {value!r}")
    return value


def _validate_linter_name(value: str) -> str:
    """Validate linter name."""
    from charmcraft import linters  # import here to avoid cyclic imports

    valid_names = [
        checker.name
        for checker in linters.CHECKERS
        if checker.check_type == linters.CheckType.LINT
    ]
    if value not in valid_names:
        raise ValueError(f"Bad lint name {value!r}")
    return value


RelativePath = craft_parts.constraints.RelativePathStr
AttributeName = Annotated[  # TODO: Turn this into a StrEnum
    str,
    pydantic.Field(strict=True),
    pydantic.BeforeValidator(_validate_attribute_name),
]
LinterName = Annotated[
    str,
    pydantic.Field(strict=True),
    pydantic.BeforeValidator(_validate_linter_name),
]



# Copyright 2023 Canonical Ltd.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# For further info, check https://github.com/canonical/charmcraft
"""File-related utilities."""
import io
import os
import pathlib
import zipfile
from _stat import S_IRGRP, S_IROTH, S_IRUSR, S_IXGRP, S_IXOTH, S_IXUSR

from craft_cli import CraftError

# handy masks for execution and reading for everybody
S_IXALL = S_IXUSR | S_IXGRP | S_IXOTH
S_IRALL = S_IRUSR | S_IRGRP | S_IROTH

PathOrString = os.PathLike | str


def make_executable(fh: io.IOBase) -> None:
    """Make open file fh executable.

    :param fh: An open file object.
    """
    fileno = fh.fileno()
    mode = os.fstat(fileno).st_mode
    mode_r = mode & S_IRALL
    mode_x = mode_r >> 2
    mode = mode | mode_x
    os.fchmod(fileno, mode)


def useful_filepath(filepath: PathOrString) -> pathlib.Path:
    """Return a valid Path with username expansion for filepath.

    CraftError is raised if filepath is not a valid file or is not readable.
    """
    filepath = pathlib.Path(filepath).expanduser()
    if not os.access(filepath, os.R_OK):
        raise CraftError(f"Cannot access {str(filepath)!r}.")
    if not filepath.is_file():
        raise CraftError(f"{str(filepath)!r} is not a file.")
    return filepath


def build_zip(zip_path: PathOrString, prime_dir: PathOrString) -> None:
    """Build a zip file from a prime directory.

    :param zip_path: The path to the output zip file
    :param prime_dir: The path to the directory to zip.
    """
    zip_path = pathlib.Path(zip_path).resolve()
    prime_dir = pathlib.Path(prime_dir).resolve()
    with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zip_file:
        # Using os.walk() because Path.walk() is only added in 3.12
        for dir_path_str, _, filenames in os.walk(prime_dir, followlinks=True):
            for filename in filenames:
                file_path = pathlib.Path(dir_path_str, filename)
                zip_file.write(file_path, file_path.relative_to(prime_dir))
